"""
selinux test utility functions.
"""

import logging
import re
import os

from avocado.utils import process
from avocado.utils import distro


ubuntu = distro.detect().name == 'Ubuntu'


class SelinuxError(Exception):

    """
    Error selinux utility functions.
    """
    pass


class SeCmdError(SelinuxError):

    """
    Error in executing cmd.
    """

    def __init__(self, cmd, detail):
        SelinuxError.__init__(self)
        self.cmd = cmd
        self.detail = detail

    def __str__(self):
        return str("Execute command %s failed.\n"
                   "Detail: %s .\n" % (self.cmd, self.detail))


class SemanageError(SelinuxError):

    """
    Error when semanage binary is not found
    """

    def __str__(self):
        return ("The semanage command is not available, "
                "please install policycoreutils "
                "or equivalent for your platform.")


class RestoreconError(SelinuxError):

    def __str__(self):
        return ("Output from the restorecon command"
                "does not match the expected format")


STATUS_LIST = ['enforcing', 'permissive', 'disabled']


def get_status(selinux_force=False):
    """
    Get the status of selinux.

    :param selinux_force: True to force selinux configuration on Ubuntu
    :return: string of status in STATUS_LIST.
    :raise SeCmdError: if execute 'getenforce' failed.
    :raise SelinuxError: if 'getenforce' command exit 0,
                    but the output is not expected.
    """
    if ubuntu and not selinux_force:
        logging.warning("Ubuntu doesn't support selinux by default")
        return 'disabled'

    cmd = 'getenforce'
    try:
        result = process.run(cmd, ignore_status=True)
    except OSError:
        raise SeCmdError(cmd, "Command not available")

    if result.exit_status:
        raise SeCmdError(cmd, result.stderr_text)

    for status in STATUS_LIST:
        if result.stdout_text.lower().count(status):
            return status
        else:
            continue

    raise SelinuxError("result of 'getenforce' (%s)is not expected."
                       % result.stdout_text)


def set_status(status, selinux_force=False):
    """
    Set status of selinux.

    :param status: status want to set selinux.
    :param selinux_force: True to force selinux configuration on Ubuntu
    :raise SelinuxError: status is not supported.
    :raise SelinuxError: need to reboot host.
    :raise SeCmdError: execute setenforce failed.
    :raise SelinuxError: cmd setenforce exit normally,
                but status of selinux is not set to expected.
    """
    if ubuntu and not selinux_force:
        logging.warning("Ubuntu doesn't support selinux by default")
        return

    if status not in STATUS_LIST:
        raise SelinuxError("Status %s is not accepted." % status)

    current_status = get_status(selinux_force)
    if status == current_status:
        return
    else:
        if current_status == "disabled" or status == "disabled":
            raise SelinuxError("Please modify /etc/selinux/config and "
                               "reboot host to set selinux to %s." % status)
        else:
            cmd = "setenforce %s" % status
            result = process.run(cmd, ignore_status=True)
            if result.exit_status:
                raise SeCmdError(cmd, result.stderr_text)
            else:
                current_status = get_status(selinux_force)
                if not status == current_status:
                    raise SelinuxError("Status of selinux is set to %s,"
                                       "but not expected %s. "
                                       % (current_status, status))
                else:
                    pass

    logging.debug("Set status of selinux to %s success.", status)


def is_disabled(selinux_force=False):
    """
    Return True if the selinux is disabled.

    :param selinux_force: True to force selinux configuration on Ubuntu
    """
    if ubuntu and not selinux_force:
        logging.warning("Ubuntu doesn't support selinux by default")
        return True

    status = get_status(selinux_force)
    if status == "disabled":
        return True
    else:
        return False


def is_not_disabled(selinux_force=False):
    """
    Return True if the selinux is not disabled.

    :param selinux_force: True to force selinux configuration on Ubuntu
    """
    if ubuntu and not selinux_force:
        logging.warning("Ubuntu doesn't support selinux by default")
        return False

    return not is_disabled(selinux_force)


def is_enforcing(selinux_force=False):
    """
    Return true if the selinux is enforcing.

    :param selinux_force: True to force selinux configuration on Ubuntu
    """
    if ubuntu and not selinux_force:
        logging.warning("Ubuntu doesn't support selinux by default")
        return False

    return (get_status(selinux_force) == "enforcing")


def is_permissive(selinux_force=False):
    """
    Return true if the selinux is permissive.

    :param selinux_force: True to force selinux configuration on Ubuntu
    """
    if ubuntu and not selinux_force:
        logging.warning("Ubuntu doesn't support selinux by default")
        return False

    return (get_status(selinux_force) == "permissive")


def get_context_from_str(context):
    """
    Get the context in a context.

    :param context: SELinux context string
    :raise SelinuxError: if there is no context in context.
    """
    context_pattern = (r"[a-z,_]*_u:[a-z,_]*_r:[a-z,_]*_t"
                       # non-greedy/non-group match on optional MLS range
                       r"(?:\:[s,\-,0-9,:[c,\,,0-9]*]*)?")
    if re.search(context_pattern, context):
        context_list = re.findall(context_pattern, context)
        return context_list[0]

    raise SelinuxError("There is no context in %s." % context)


def get_type_from_context(context):
    """
    Return just the type component of a full context string

    :param context: SELinux context string
    :return: Type component of SELinux context string
    """
    # Raise exception if not a context string
    get_context_from_str(context)
    type_pattern = (r"[a-z,_]*_u:[a-z,_]*_r:([a-z,_]*_t)"
                    r"(?:\:[s,\-,0-9,:[c,\,,0-9]*]*)?")
    return re.search(type_pattern, context).group(1)


def get_context_of_file(filename, selinux_force=False):
    """
    Get the context of file.

    :param filename: filename for the context to be get
    :param selinux_force: True to force selinux configuration on Ubuntu
    :raise SeCmdError: if execute 'getfattr' failed.
    """
    if ubuntu and not selinux_force:
        logging.warning("Ubuntu doesn't support selinux by default")
        return

    # More direct than scraping 'ls' output.
    cmd = "getfattr --name security.selinux %s" % filename
    result = process.run(cmd, ignore_status=True)
    if result.exit_status:
        raise SeCmdError(cmd, result.stderr_text)

    output = result.stdout_text
    return get_context_from_str(output)


def set_context_of_file(filename, context, selinux_force=False):
    """
    Set context of file.

    :param filename: filename for the context to be set
    :param context: new value of the extended context attribute
    :param selinux_force: True to force selinux configuration on Ubuntu
    :raise SeCmdError: if failed to execute chcon.
    :raise SelinuxError: if command chcon execute
                        normally, but the context of
                        file is not setted to context.
    """
    if ubuntu and not selinux_force:
        logging.warning("Ubuntu doesn't support selinux by default")
        return

    context = context.strip()
    # setfattr used for consistency with getfattr use above
    cmd = ("setfattr --name security.selinux --value \"%s\" %s"
           % (context, filename))
    result = process.run(cmd, ignore_status=True)
    if result.exit_status:
        raise SeCmdError(cmd, result.stdout_text)

    context_result = get_context_of_file(filename)
    if not context == context_result:
        raise SelinuxError("Context of %s after chcon is %s, "
                           "but not expected %s."
                           % (filename, context_result, context))

    logging.debug("Set context of %s success.", filename)


def check_context_of_file(filename, label, selinux_force=False):
    """
    Check for label in the context of given filename.

    :param filename: filename for which context to be retrieved
    :param label: label to be checked in the context
    :param selinux_force: True to force selinux configuration on Ubuntu
    """
    se_label = get_context_of_file(filename, selinux_force)
    if se_label is not None:
        logging.debug("Context of shared filename '%s' is '%s'" %
                      (filename, se_label))
        if label not in se_label:
            return False
    else:
        logging.warning("Context of shared filename '%s' is None" % filename)
        return False
    return True


def get_context_of_process(pid):
    """
    Get context of process.
    """
    attr_filepath = "/proc/%s/attr/current" % pid

    attr_file = open(attr_filepath)

    output = attr_file.read()
    return get_context_from_str(output)

# Force uniform handling if semanage not found (used in unittests)


def _no_semanage(cmdresult):
    if cmdresult.exit_status == 127:
        if cmdresult.stdout_text.lower().count('command not found'):
            raise SemanageError()


def get_defcon(local=False, selinux_force=False):
    """
    Return list of dictionaries containing SELinux default file context types

    :param local: Only return locally modified default contexts
    :param selinux_force: True to force selinux configuration on Ubuntu
    :return: list of dictionaries of default context attributes
    """
    if ubuntu and not selinux_force:
        logging.warning("Ubuntu doesn't support selinux by default")
        return

    if local:
        result = process.run("semanage fcontext --list -C", ignore_status=True)
    else:
        result = process.run("semanage fcontext --list", ignore_status=True)
    _no_semanage(result)
    if result.exit_status != 0:
        raise SeCmdError('semanage', result.stderr_text)
    result_list = result.stdout_text.strip().split('\n')
    # Need to process top-down instead of bottom-up
    result_list.reverse()
    first_line = result_list.pop()
    # First column name has a space in it
    column_names = [name.strip().lower().replace(' ', '_')
                    for name in first_line.split('  ')
                    if len(name) > 0]
    # Shorten first column name
    column_names[0] = column_names[0].replace("selinux_", "")
    fcontexts = []
    for line in result_list:
        if len(line) < 1:  # skip blank lines
            continue
        column_data = [name.strip()
                       for name in line.split('  ')
                       if len(name) > 0]
        # Enumerating data raises exception if no column_names match
        fcontext = dict([(column_names[idx], data)
                         for idx, data in enumerate(column_data)])
        # find/set functions only accept type, not full context string
        fcontext['context'] = get_type_from_context(fcontext['context'])
        fcontexts.append(fcontext)
    return fcontexts


def find_defcon_idx(defcon, pathname):
    """
    Returns the index into defcon where pathname matches or None
    """
    # Default context path regexes only work on canonical paths
    pathname = os.path.realpath(pathname)
    for default_context in defcon:
        if bool(re.search(default_context['fcontext'], pathname)):
            return defcon.index(default_context)
    return None


def find_defcon(defcon, pathname):
    """
    Returns the context type of first match to pathname or None
    """
    # Default context path regexes only work on canonical paths
    pathname = os.path.realpath(pathname)
    idx = find_defcon_idx(defcon, pathname)
    if idx is not None:
        return get_type_from_context(defcon[idx]['context'])
    else:
        return None


def find_pathregex(defcon, pathname):
    """
    Returns the regular expression in defcon matching pathname
    """
    # Default context path regexes only work on canonical paths
    pathname = os.path.realpath(pathname)
    idx = find_defcon_idx(defcon, pathname)
    if idx is not None:
        return defcon[idx]['fcontext']
    else:
        return None


def set_defcon(context_type, pathregex, context_range=None, selinux_force=False):
    """
    Set the default context of a file/path in local SELinux policy

    :param context_type: The selinux context (only type is used)
    :param pathregex: Pathname regex e.g. r"/foo/bar/baz(/.*)?"
    :param context_range: MLS/MCS Security Range e.g. s0:c87,c520
    :param selinux_force: True to force selinux configuration on Ubuntu
    :raise SelinuxError: if semanage command not found
    :raise SeCmdError: if semanage exits non-zero
    """
    if ubuntu and not selinux_force:
        logging.warning("Ubuntu doesn't support selinux by default")
        return

    cmd = "semanage fcontext --add"
    if context_type:
        cmd += ' -t %s' % context_type
    if context_range:
        cmd += ' -r %s' % context_range
    if pathregex:
        cmd += ' "%s"' % pathregex
    result = process.run(cmd, ignore_status=True)
    result.stdout = result.stdout_text
    result.stderr = result.stderr_text
    _no_semanage(result)
    if result.exit_status != 0:
        raise SeCmdError(cmd, result.stderr_text)


def del_defcon(context_type, pathregex, selinux_force=False):
    """
    Remove the default local SELinux policy type for a file/path

    :param context: The selinux context (only type is used)
    :pramm pathregex: Pathname regex e.g. r"/foo/bar/baz(/.*)?"
    :param selinux_force: True to force selinux configuration on Ubuntu
    :raise SelinuxError: if semanage command not found
    :raise SeCmdError: if semanage exits non-zero
    """
    if ubuntu and not selinux_force:
        logging.warning("Ubuntu doesn't support selinux by default")
        return

    cmd = ("semanage fcontext --delete -t %s '%s'" % (context_type, pathregex))
    result = process.run(cmd, ignore_status=True)
    result.stdout = result.stdout_text
    result.stderr = result.stderr_text
    _no_semanage(result)
    if result.exit_status != 0:
        raise SeCmdError(cmd, result.stderr_text)

# Process pathname/dirdesc in uniform way for all defcon functions + unittests


def _run_restorecon(pathname, dirdesc, readonly=True, force=False, selinux_force=False):
    """
    Use restorecon to restore selinux context for file

    :param pathname: Absolute path to file, directory, or symlink
    :param dirdesc: True to descend into sub-directories
    :param readonly: True to passive check and don't change any file labels
    :param force: True to force reset of context to match file_context
    :param selinux_force: True to force selinux configuration on Ubuntu
    """
    if ubuntu and not selinux_force:
        logging.warning("Ubuntu doesn't support selinux by default")
        return 0

    cmd = 'restorecon -v'
    if dirdesc:
        cmd += 'R'
    if readonly:
        cmd += 'n'
    if force:
        cmd += 'F'
    cmd += ' "%s"' % pathname
    # Always returns 0, even if contexts wrong
    return process.run(cmd).stdout_text.strip()


def verify_defcon(pathname, dirdesc=False, readonly=True, forcedesc=False, selinux_force=False):
    """
    Verify contexts of pathspec (and/or below, if dirdesc) match default

    :param pathname: Absolute path to file, directory, or symlink
    :param dirdesc: True to descend into sub-directories
    :param readonly: True to passive check and don't change any file labels
    :param forcedesc: True to force a replacement of the entire context
    :param selinux_force: True to force selinux configuration on Ubuntu
    :return: True if all components match default contexts
    :note: By default DOES NOT follow symlinks
    """
    if ubuntu and not selinux_force:
        logging.warning("Ubuntu doesn't support selinux by default")
        return False
    # Default context path regexes only work on canonical paths
    changes = _run_restorecon(pathname, dirdesc,
                              readonly=readonly, force=forcedesc,
                              selinux_force=selinux_force)
    if changes.count('restorecon reset'):
        return False
    else:
        return True


# Provide uniform formatting for diff and apply functions

def _format_changes(changes):
    result = []
    if changes:  # Empty string or None - return empty list
        # Could be many changes, need efficient line searching
        regex = re.compile('^restorecon reset (.+) context (.+)->(.+)')
        for change_line in changes.split('\n'):
            mobj = regex.search(change_line)
            if mobj is None:
                raise RestoreconError()
            pathname = mobj.group(1)
            from_con = mobj.group(2)
            to_con = mobj.group(3)
            result.append((pathname, from_con, to_con))
    return result


def diff_defcon(pathname, dirdesc=False, selinux_force=False):
    """
    Return a list of tuple(pathname, from, to) for current & default contexts

    :param pathname: Absolute path to file, directory, or symlink
    :param dirdesc: True to descend into sub-directories
    :return: List of tuple(pathname, from context, to context)
    """
    return _format_changes(_run_restorecon(pathname, dirdesc,
                           selinux_force=selinux_force))


def apply_defcon(pathname, dirdesc=False, selinux_force=False):
    """
    Apply default contexts to pathname, possibly descending into sub-dirs also.

    :param pathname: Absolute path to file, directory, or symlink
    :param dirdesc: True to descend into sub-directories
    :return: List of changes applied tuple(pathname, from context, to context)
    """
    return _format_changes(_run_restorecon(pathname, dirdesc, readonly=False,
                                           selinux_force=selinux_force))


def transmogrify_usr_local(pathregex):
    """
    Replace usr/local/something with usr/(local/)?something
    """
    # Whoa! don't mess with short path regex's
    if len(pathregex) < 3:
        return pathregex
    if pathregex.count('usr/local'):
        pathregex = pathregex.replace('usr/local/', r'usr/(local/)?')
    return pathregex


def transmogrify_sub_dirs(pathregex):
    """
    Append '(/.*)?' regex to end of pathregex to optionally match all subdirs
    """
    # Whoa! don't mess with short path regex's
    if len(pathregex) < 3:
        return pathregex
    # Doesn't work with path having trailing slash
    if pathregex.endswith('/'):
        pathregex = pathregex[0:-1]
    return pathregex + r'(/.*)?'



"""
Module simplifying manipulation of sysinfo XML
"""

from virttest.libvirt_xml import base


class SysinfoXML(base.LibvirtXMLBase):

    """
    Handler of libvirt sysinfo xml.
    """

    # TODO: Add more __slots__, accessors and functions to get some useful
    # stats

    __slots__ = []

    def __init__(self, virsh_instance=base.virsh):
        """
        Initialize new instance with empty XML
        """
        super(SysinfoXML, self).__init__(virsh_instance=virsh_instance)
        self.xml = u"<sysinfo></sysinfo>"

    def get_all_processors(self):
        """
        Get all processors dict with entry name as key.

        :return: all processors dict with entry name as key
        """
        processor_dict = {}
        processor_nodes = self.xmltreefile.findall('processor')
        for i in range(len(processor_nodes)):
            temp_dict = {}
            entry_nodes = processor_nodes[i].getchildren()
            if entry_nodes:
                for entry in entry_nodes:
                    entry_attr = dict(list(entry.items()))
                    if 'name' in entry_attr:
                        temp_dict[entry_attr['name']] = entry.text
                processor_dict[i] = temp_dict

        return processor_dict



#!/usr/bin/python
import socket
import struct
import os
import signal
import sys
# -*- coding: utf-8 -*-

"""
Script used to join machine into multicast groups.

:author Amos Kong <akong@redhat.com>
"""

if __name__ == "__main__":
    if len(sys.argv) < 4:
        print("""%s [mgroup_count] [prefix] [suffix]
        mgroup_count: count of multicast addresses
        prefix: multicast address prefix
        suffix: multicast address suffix""" % sys.argv[0])
        sys.exit()

    mgroup_count = int(sys.argv[1])
    prefix = sys.argv[2]
    suffix = int(sys.argv[3])

    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    for i in range(mgroup_count):
        mcast = prefix + "." + str(suffix + i)
        try:
            mreq = struct.pack("4sl", socket.inet_aton(mcast),
                               socket.INADDR_ANY)
            s.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
        except Exception:
            s.close()
            print("Could not join multicast: %s" % mcast)
            raise

    print("join_mcast_pid:%s" % os.getpid())
    os.kill(os.getpid(), signal.SIGSTOP)
    s.close()



"""
input device support class(es)

http://libvirt.org/formatdomain.html#elementsInput
"""

from virttest.libvirt_xml import accessors
from virttest.libvirt_xml.devices import base, librarian


class Input(base.TypedDeviceBase):
    __slots__ = ('input_bus', 'model', 'address', 'source_evdev')

    def __init__(self, type_name, virsh_instance=base.base.virsh):
        super(Input, self).__init__(device_tag='input',
                                    type_name=type_name,
                                    virsh_instance=virsh_instance)
        accessors.XMLAttribute(property_name="input_bus",
                               libvirtxml=self,
                               forbidden=None,
                               parent_xpath='/',
                               tag_name='input',
                               attribute='bus')
        accessors.XMLAttribute(property_name="model",
                               libvirtxml=self,
                               parent_xpath='/',
                               tag_name='input',
                               attribute='model')
        accessors.XMLElementNest('address', self, parent_xpath='/',
                                 tag_name='address', subclass=self.Address,
                                 subclass_dargs={'type_name': 'usb',
                                                 'virsh_instance': virsh_instance})
        accessors.XMLAttribute('source_evdev', self, parent_xpath='/',
                               tag_name='source', attribute='evdev')
    # For convenience
    Address = librarian.get('address')

    def new_input_address(self, type_name='usb', **dargs):
        """
        Return a new input Address instance and set properties from dargs
        """
        new_one = self.Address(type_name=type_name, virsh_instance=self.virsh)
        for key, value in list(dargs.items()):
            setattr(new_one, key, value)
        return new_one



"""
Virtualization test utility functions.

:copyright: 2008-2009 Red Hat Inc.
"""
import multiprocessing.pool
import functools


def timeout(timeout):
    """Timeout decorator, parameter in seconds."""
    def timeout_decorator(func):
        """Wrap the original function."""
        @functools.wraps(func)
        def func_wrapper(*args, **kwargs):
            """Closure for function."""
            pool = multiprocessing.pool.ThreadPool(processes=1)
            async_result = pool.apply_async(func, args, kwargs)
            # raises a TimeoutError if execution exceeds timeout
            return async_result.get(timeout)
        return func_wrapper
    return timeout_decorator



"""
Windows drive utilities
"""

from . import wmic
from virttest import utils_misc


def _logical_disks(session, cond=None, props=None):
    cmd = wmic.make_query("LogicalDisk", cond, props,
                          get_swch=wmic.FMT_TYPE_LIST)
    out = utils_misc.wait_for(lambda: wmic.parse_list(session.cmd(cmd,
                              timeout=120)), 240)
    return out if out else []


def get_hard_drive_letter(session, label):
    """
    Get hard drive's letter by the given label.

    :param session: Session object.
    :param label: Label pattern string.

    :return: Hard drive's letter if found, otherwise `None`.
    """
    cond = "VolumeName like '%s'" % label
    try:
        return _logical_disks(session, cond=cond, props=["DeviceID"])[0]
    except IndexError:
        return None


def get_floppy_drives_letter(session):
    """
    Get all the floppy drives' letter.

    :param session: Session object.

    :return: Floppy drives' letter.
    """
    cond = "MediaType!=0 AND MediaType!=11 AND MediaType!=12"
    return _logical_disks(session, cond=cond, props=["DeviceID"])


def rescan_disks(session):
    """
    Rescan disks in windows guest.

    :param session: Session object.
    """
    script_path = r"%TEMP%\rescan.dp"
    rescan_cmd = "echo rescan > {0} && diskpart /s {0}"
    session.cmd(rescan_cmd.format(script_path))


def extend_volume(session, vol_id, size=None):
    """
    Extend a volume in windows guest.

    :param session: Session object.
    :param vol_id: Drive letter or Volume number.
    :param size: Default extend the volume to maximum available size,
                 if size is specified, extend the volume to size.
                 The default unit of size is M.
    """
    script_path = r"%TEMP%\extend_{0}.dp".format(vol_id)
    extend_cmd = 'echo select volume %s > {0} && ' % vol_id
    if not size:
        extend_cmd += 'echo extend >> {0} && diskpart /s {0}'
    else:
        extend_cmd += 'echo extend desired=%s >> {0} ' % size
        extend_cmd += '&& diskpart /s {0}'
    session.cmd(extend_cmd.format(script_path))


def shrink_volume(session, vol_id, size):
    """
    Shrink a volume in windows guest.

    :param session: Session object.
    :param vol_id: Drive letter or Volume number.
    :param size: Desired decrease size. The default unit of size is M.
    """
    script_path = r"%TEMP%\shrink_{0}.dp".format(vol_id)
    shrink_cmd = 'echo select volume %s > {0} && ' % vol_id
    shrink_cmd += 'echo shrink desired=%s >> {0} ' % size
    shrink_cmd += '&& diskpart /s {0}'
    session.cmd(shrink_cmd.format(script_path))



#!/usr/bin/env python

'''
Created on Dec 6, 2013

:author: jzupka, astepano
:contact: Andrei Stepanov <astepano@redhat.com>
'''

from __future__ import division
import sys
import time
import inspect

from virttest.remote_commander import remote_interface
from virttest.remote_commander import messenger


def getsource(obj):
    return inspect.getsource(obj)


def wait_timeout(timeout):
    if timeout is None:
        while 1:
            yield True
    else:
        end_time = time.time() + timeout
        while time.time() < end_time:
            yield True


class CmdMaster(object):

    """
    Representation of BaseCmd on master side.
    """

    def __init__(self, commander, name, *args, **kargs):
        """
        :params commander: Commander from which was command started.
        :params name: Name parsed to string representation
        :type name: [str, str, str]
        :parmas args: list to arguments to cmd.
        :type args: []
        :params kargs: {}
        """
        self._basecmd = remote_interface.BaseCmd(name, *args, **kargs)
        self.commander = commander
        self._stdout = ""
        self._stderr = ""
        self._results_cnt = 0
        self._stdout_cnt = 0
        self._stderr_cnt = 0
        self.timeout = kargs.get('timeout', None)

    def getbasecmd(self):
        """
        Property basecmd getter
        """
        self._results_cnt = 0
        return self._basecmd

    def setbasecmd(self, value):
        """
        Property basecmd setter _resuls_cnt identify if value was change from
        last reading.
        """

        self._basecmd = value
        self._results_cnt += 1

    basecmd = property(getbasecmd, setbasecmd)

    def getstdout(self):
        """
        Property stdout getter
        """
        self._stdout_cnt = 0
        return self._stdout

    def setstdout(self, value):
        """
        Property stdout setter _stdout_cnt identify if value was change from
        last reading.
        """
        self._stdout = value
        self._stdout_cnt += 1

    stdout = property(getstdout, setstdout)

    def getstderr(self):
        """
        Property stderr getter
        """
        self._stderr_cnt = 0
        return self._stderr

    def setstderr(self, value):
        """
        Property stderr setter _stderr_cnt identify if value was change from
        last reading.
        """
        self._stderr = value
        self._stderr_cnt += 1

    stderr = property(getstderr, setstderr)

    def send_stdin(self, msg):
        """
        Send data to stdin
        """
        self.commander.manage.send_msg(msg, self.basecmd.cmd_id)

    def wait(self):
        """
        Wait until command return results.
        """
        return self.commander.wait(self)

    def wait_response(self):
        """
        Wait until command return any cmd.
        """
        self.commander.wait_response(self)

    def __getattr__(self, name):
        """
        Shortcut to encapsulated basecmd.
        """
        if name in ["__getstate__", "__setstate__", "__slots__"]:
            raise AttributeError()
        return getattr(self.basecmd, name)

    def set_commander(self, commander):
        """
        For nohup commands it allows connect cmd to new created commander.

        """
        self.commander = commander
        if self not in commander.cmds:
            commander.cmds[self.cmd_id] = self
        self.commander.manage.register_cmd(self.basecmd,
                                           remote_interface.BaseCmd.single_cmd_id)


class CmdEncapsulation(object):

    """
    Class parse command name   cmd.nohup.shell -> ["nohup", "shell"]
    """

    def __init__(self, master, obj_name, name):
        self.master = master
        if obj_name is None:
            self.name = [name]
        else:
            self.name = obj_name + [name]
        self.cmd = None

    def __getattr__(self, name):
        return CmdEncapsulation(self.master, self.name, name)

    def __call__(self, *args, **kargs):
        """
        Call commander with specific command.
        """
        self.cmd = CmdMaster(self.master, self.name, *args, **kargs)
        return self.master.cmd(self.cmd)


class CmdTimeout(remote_interface.MessengerError):

    """
    Raised when waiting for cmd exceeds time define by timeout.
    """

    def __init__(self, msg):
        super(CmdTimeout, self).__init__(msg)

    def __str__(self):
        return "Commander Timeout %s" % (self.msg)


class Commander(object):

    """
    Commander representation for transfer over network.
    """
    __slots__ = []


class CommanderMaster(messenger.Messenger):

    """
    Class commander master is responsible for communication with commander
    slave. It invoke commands to slave part and receive messages from them.
    For communication is used only stdin and stdout which are streams from
    slave part.
    """

    def __init__(self, stdin, stdout, debug=False):
        """
        :type stdin: IOWrapper with implemented write function.
        :type stout: IOWrapper with implemented read function.
        """
        super(CommanderMaster, self).__init__(stdin, stdout)
        self.cmds = {}
        self.debug = debug
        self.responder = None

        self.flush_stdin()
        self.write_msg("start")
        succ, msg = self.read_msg()
        if not succ or msg != "Started":
            raise remote_interface.CommanderError("Remote commander"
                                                  " not started.")

    def set_responder(self, responder):
        """
        :warning: Users's helper becomes a part of a remote_commander.
            remote_commander is treated as a session. All sessions are part of
            BaseVM object. If you provide unpickable helper then your BaseVM
            also becomes unpickable. You can catch something like::

                File "/usr/lib64/python2.7/copy_reg.py", line 74, in _reduce_ex
                getstate = self.__getstate__
                KeyError: '__getstate__'

            For more info see::

                https://github.com/avocado-framework/avocado-vt/issues/455

        """
        self.responder = responder

    def close(self):
        try:
            self.manage.exit()
        except Exception:
            pass
        super(CommanderMaster, self).close()

    def __getattr__(self, name):
        """
        Start parsing unknown attribute in cmd.
        """
        if name in ["__getstate__", "__setstate__", "__slots__"]:
            raise AttributeError()
        return CmdEncapsulation(self, None, name)

    def __deepcopy__(self, memo):
        """
        Replace deepcopy by substituting by network Commander version.
        """
        result = Commander.__new__(Commander)
        memo[id(self)] = result
        return result

    def listen_streams(self, cmd):
        """
        Listen on all streams included in Commander commands.
        """
        if isinstance(cmd, remote_interface.StdStream):
            if (self.debug):
                print(cmd.msg)
            if cmd.isCmdMsg():
                if isinstance(cmd, remote_interface.StdOut):
                    self.cmds[cmd.cmd_id].stdout += cmd.msg
                elif isinstance(cmd, remote_interface.StdErr):
                    self.cmds[cmd.cmd_id].stderr += cmd.msg
            else:
                if isinstance(cmd, remote_interface.StdOut):
                    sys.stdout.write(cmd.msg)
                elif isinstance(cmd, remote_interface.StdErr):
                    sys.stderr.write(cmd.msg)

    def listen_errors(self, cmd):
        """
        Listen for errors raised from slave part of commander.
        """
        if isinstance(cmd, (Exception, remote_interface.CommanderError,
                            remote_interface.MessengerError)):
            raise cmd

    def listen_queries(self, cmd):
        """Manage queries from slave side.
        """
        if isinstance(cmd, remote_interface.CmdQuery):
            assert self.responder  # Should be set by set_responder()
            data = self.responder(*cmd.args, **cmd.kargs)
            reply = remote_interface.CmdRespond(data)
            self.write_msg(reply)

    def listen_cmds(self, cmd):
        """
        Manage basecmds from slave side.
        """
        if isinstance(cmd, remote_interface.BaseCmd):
            if (self.debug):
                print(cmd.func, cmd.results, cmd._finished)

            if isinstance(cmd.results, Exception):
                raise cmd.results
            if cmd.cmd_id in self.cmds:
                self.cmds[cmd.cmd_id].basecmd.update(cmd)
                self.cmds[cmd.cmd_id].basecmd.update_cmd_hash(cmd)

    def listen_messenger(self, timeout=60):
        """
        Wait for msg from slave side and take care about them.
        """
        succ, r_cmd = self.read_msg(timeout)
        if succ is None:
            return r_cmd
        if not succ:
            raise remote_interface.CommanderError("Remote process died.")

        self.listen_errors(r_cmd)
        self.listen_streams(r_cmd)
        self.listen_cmds(r_cmd)
        self.listen_queries(r_cmd)
        return r_cmd

    def cmd(self, cmd):
        """
        Invoke command on client side.
        """
        self.cmds[cmd.basecmd.cmd_id] = cmd
        self.write_msg(cmd.basecmd)
        while (1):
            if cmd.basecmd.func[0] not in ["async", "nohup"]:
                # If not async wait for finish.
                self.wait(cmd)
            else:
                ancmd = self.wait_response(cmd)
                cmd.update_cmd_hash(ancmd)
            return cmd

    def wait(self, cmd):
        """
        Wait until command return results.
        """
        if cmd.cmd_id not in self.cmds:
            return cmd
        m_cmd = self.cmds[cmd.cmd_id]
        if m_cmd.is_finished():
            return m_cmd

        r_cmd = None

        time_step = None
        timeout = cmd.timeout
        if timeout is not None:
            time_step = timeout / 10.0
        w = wait_timeout(timeout)
        for _ in w:
            r_cmd = self.listen_messenger(time_step)
            if isinstance(r_cmd, remote_interface.BaseCmd):
                if (self.debug):
                    print(m_cmd._stdout)
                if r_cmd is not None and r_cmd == m_cmd.basecmd:
                    # If command which we waiting for.
                    if r_cmd.is_finished():
                        del self.cmds[m_cmd.basecmd.cmd_id]
                        m_cmd.basecmd.update(r_cmd)
                        return m_cmd
                    m_cmd.basecmd.update(r_cmd)
                    m_cmd.basecmd.update_cmd_hash(r_cmd)

        if r_cmd is None:
            raise CmdTimeout("%ss during %s" % (timeout, str(cmd)))

    def wait_response(self, cmd):
        """
        Wait until command return any cmd.
        """
        if cmd.cmd_id not in self.cmds:
            return cmd
        if cmd.is_finished() or cmd._stdout_cnt or cmd._stderr_cnt:
            return cmd
        m_cmd = self.cmds[cmd.cmd_id]

        r_cmd = None

        time_step = None
        timeout = cmd.timeout
        if timeout is not None:
            time_step = timeout / 10.0
        w = wait_timeout(timeout)
        while (next(w)):
            r_cmd = self.listen_messenger(time_step)
            if r_cmd is not None and r_cmd == m_cmd.basecmd:
                return m_cmd

        if r_cmd is None:
            raise CmdTimeout(timeout)



"""
CEPH Support
This file has the functions that helps
* To create rbd pool
* To create rbd namespace
* To map/unmap rbd pool
* To mount/umount cephfs to localhost
* To return rbd uri which can be used as disk image file path.
"""

import logging
import os
import re

from avocado.utils import process

from virttest import utils_numeric
from virttest import error_context
from virttest import utils_misc


class CephError(Exception):
    pass


@error_context.context_aware
def rbd_image_create(ceph_monitor, rbd_pool_name, rbd_image_name,
                     rbd_image_size, force_create=False, ceph_conf=None,
                     keyfile=None, rbd_namespace_name=None):
    """
    Create a rbd image.
    :params ceph_monitor: The specified monitor to connect to
    :params rbd_pool_name: The name of rbd pool
    :params rbd_namespace_name: The name of rbd namespace
    :params rbd_image_name: The name of rbd image
    :params rbd_image_size: The size of rbd image
    :params force_create: Force create the image or not
    :params ceph_conf: The path to the ceph configuration file
    :params keyfile: The path to the ceph keyring configuration file
    """
    create_image = True
    try:
        int(rbd_image_size)
        compare_str = rbd_image_size
    except ValueError:
        compare_str = utils_numeric.normalize_data_size(rbd_image_size, 'M')

    if rbd_image_exist(ceph_monitor, rbd_pool_name, rbd_image_name,
                       ceph_conf, keyfile, rbd_namespace_name):
        create_image = False
        image_info = rbd_image_info(ceph_monitor, rbd_pool_name,
                                    rbd_image_name, ceph_conf, keyfile,
                                    rbd_namespace_name)
        if image_info['size'] != compare_str or force_create:
            rbd_image_rm(ceph_monitor, rbd_pool_name, rbd_image_name,
                         ceph_conf, keyfile, rbd_namespace_name)
            create_image = True

    if create_image:
        cmd = "rbd {opts} create {pool}/{namespace}{image} {size} {keyring}"
        c_opt = '-c %s' % ceph_conf if ceph_conf else ''
        m_opt = '-m %s' % ceph_monitor if ceph_monitor else ''
        opts = m_opt + ' ' + c_opt
        namespace = '%s/' % rbd_namespace_name if rbd_namespace_name else ''
        size = '-s %d' % utils_numeric.align_value(compare_str, 1)
        keyring = '--keyring %s' % keyfile if keyfile else ''
        cmd = cmd.format(opts=opts, pool=rbd_pool_name, namespace=namespace,
                         image=rbd_image_name, size=size, keyring=keyring)
        process.system(cmd, verbose=True)
    else:
        logging.debug("Image already exist skip the create.")


@error_context.context_aware
def rbd_image_rm(ceph_monitor, rbd_pool_name, rbd_image_name,
                 ceph_conf=None, keyfile=None, rbd_namespace_name=None):
    """
    Remove a rbd image
    :params ceph_monitor: The specified monitor to connect to
    :params rbd_pool_name: The name of rbd pool
    :params rbd_namespace_name: The namespace of rbd image
    :params rbd_image_name: The name of rbd image
    :params ceph_conf: The path to the ceph configuration file
    :params keyfile: The path to the ceph keyring configuration file
    """
    if rbd_image_exist(ceph_monitor, rbd_pool_name, rbd_image_name, ceph_conf,
                       keyfile, rbd_namespace_name):
        cmd = "rbd {opts} rm {pool}/{namespace}{image} {keyring}"
        c_opt = '-c %s' % ceph_conf if ceph_conf else ''
        m_opt = '-m %s' % ceph_monitor if ceph_monitor else ''
        opts = m_opt + ' ' + c_opt
        namespace = '%s/' % rbd_namespace_name if rbd_namespace_name else ''
        keyring = '--keyring %s' % keyfile if keyfile else ''
        cmd = cmd.format(opts=opts, pool=rbd_pool_name, namespace=namespace,
                         image=rbd_image_name, keyring=keyring)
        process.run(cmd, verbose=True)
    else:
        logging.debug("Image not exist, skip to remove it.")


@error_context.context_aware
def rbd_image_exist(ceph_monitor, rbd_pool_name,
                    rbd_image_name, ceph_conf=None, keyfile=None,
                    rbd_namespace_name=None):
    """
    Check if rbd image is exist
    :params ceph_monitor: The specified monitor to connect to
    :params rbd_pool_name: The name of rbd pool
    :params rbd_namespace_name: The name of rbd namespace
    :params rbd_image_name: The name of rbd image
    :params ceph_conf: The path to the ceph configuration file
    :params keyfile: The path to the ceph keyring configuration file
    """
    cmd = "rbd {opts} ls {pool}{namespace} {keyring}"
    c_opt = '-c %s' % ceph_conf if ceph_conf else ''
    m_opt = '-m %s' % ceph_monitor if ceph_monitor else ''
    opts = m_opt + ' ' + c_opt
    keyring = '--keyring %s' % keyfile if keyfile else ''
    namespace = '/%s' % rbd_namespace_name if rbd_namespace_name else ''
    cmd = cmd.format(opts=opts, pool=rbd_pool_name, keyring=keyring,
                     namespace=namespace)

    output = process.run(cmd, ignore_status=True,
                         verbose=True).stdout_text

    logging.debug("Response from rbd ls command is: %s" % output)

    return (rbd_image_name.strip() in output.splitlines())


@error_context.context_aware
def rbd_image_info(ceph_monitor, rbd_pool_name, rbd_image_name,
                   ceph_conf=None, keyfile=None, rbd_namespace_name=None):
    """
    Get information of a rbd image
    :params ceph_monitor: The specified monitor to connect to
    :params rbd_pool_name: The name of rbd pool
    :params rbd_namespace_name: The name of rbd namespace
    :params rbd_image_name: The name of rbd image
    :params ceph_conf: The path to the ceph configuration file
    :params keyfile: The path to the ceph keyring configuration file
    """
    cmd = "rbd {opts} info {pool}/{namespace}{image} {keyring}"
    c_opt = '-c %s' % ceph_conf if ceph_conf else ''
    m_opt = '-m %s' % ceph_monitor if ceph_monitor else ''
    opts = m_opt + ' ' + c_opt
    namespace = '%s/' % rbd_namespace_name if rbd_namespace_name else ''
    keyring = '--keyring %s' % keyfile if keyfile else ''
    cmd = cmd.format(opts=opts, pool=rbd_pool_name, image=rbd_image_name,
                     namespace=namespace, keyring=keyring)
    output = process.run(cmd).stdout_text
    info_pattern = "rbd image \'%s\':.*?$" % rbd_image_name
    rbd_image_info_str = re.findall(info_pattern, output, re.S)[0]

    rbd_image_info = {}
    for rbd_image_line in rbd_image_info_str.splitlines():
        if ":" not in rbd_image_line:
            if "size" in rbd_image_line:
                size_str = re.findall("size\s+(\d+\s+\w+)\s+",
                                      rbd_image_line)[0]
                size = utils_numeric.normalize_data_size(size_str, 'M')
                rbd_image_info['size'] = size
            if "order" in rbd_image_line:
                rbd_image_info['order'] = int(re.findall("order\s+(\d+)",
                                                         rbd_image_line))
        else:
            tmp_str = rbd_image_line.strip().split(":")
            rbd_image_info[tmp_str[0]] = tmp_str[1]
    return rbd_image_info


@error_context.context_aware
def rbd_image_map(ceph_monitor, rbd_pool_name, rbd_image_name):
    """
    Maps the specified image to a block device via rbd kernel module
    :params ceph_monitor: The specified monitor to connect to
    :params rbd_pool_name: The name of rbd pool
    :params rbd_image_name: The name of rbd image
    """
    cmd = "rbd map %s --pool %s -m %s" % (rbd_image_name, rbd_pool_name,
                                          ceph_monitor)
    output = process.run(cmd, verbose=True).stdout_text
    if os.path.exist(os.path.join("/dev/rbd", rbd_pool_name, rbd_image_name)):
        return os.path.join("/dev/rbd", rbd_pool_name, rbd_image_name)
    else:
        logging.debug("Failed to map image to local: %s" % output)
        return None


@error_context.context_aware
def rbd_image_unmap(rbd_pool_name, rbd_image_name):
    """
    Unmaps the block device that was mapped via the rbd kernel module
    :params rbd_pool_name: The name of rbd pool
    :params rbd_image_name: The name of rbd image
    """
    cmd = "rbd unmap /dev/rbd/%s/%s" % (rbd_pool_name, rbd_image_name)
    output = process.run(cmd, verbose=True).stdout_text
    if os.path.exist(os.path.join("/dev/rbd", rbd_pool_name, rbd_image_name)):
        logging.debug("Failed to unmap image from local: %s" % output)


@error_context.context_aware
def get_image_filename(ceph_monitor, rbd_pool_name, rbd_image_name,
                       ceph_conf=None, rbd_namespace_name=None):
    """
    Configuration has already been configured in the conf file
    Return the rbd image file name
    :params ceph_monitor: The specified monitor to connect to
    :params rbd_pool_name: The name of rbd pool
    :params rbd_namespace_name: The name of rbd namespace
    :params rbd_image_name: The name of rbd image
    :params ceph_conf: The path to the ceph configuration file
    """
    uri = 'rbd:{pool}/{namespace}{image}{opts}'
    conf_opt = ':conf=%s' % ceph_conf if ceph_conf else ''
    mon_opt = ':mon_host=%s' % ceph_monitor if ceph_monitor else ''
    opts = conf_opt + mon_opt
    namespace = '%s/' % rbd_namespace_name if rbd_namespace_name else ''

    return uri.format(pool=rbd_pool_name, namespace=namespace,
                      image=rbd_image_name, opts=opts)


@error_context.context_aware
def create_config_file(ceph_monitor):
    """
    Create an ceph config file when the config file is not exist
    :params ceph_monitor: The specified monitor to connect to
    """
    ceph_dir = "/etc/ceph"
    ceph_cfg = os.path.join(ceph_dir, "ceph.conf")
    if not os.path.exists(ceph_dir):
        os.makedirs(ceph_dir)
    if not os.path.exists(ceph_cfg):
        with open(ceph_cfg, 'w+') as f:
            f.write('mon_host = %s' % ceph_monitor)
    return ceph_cfg


@error_context.context_aware
def cephfs_mount(ceph_uri, mount_point, options=None, verbose=False, session=None):
    """
    Mount ceph volume to mountpoint.

    :param ceph_uri: str to express ceph uri, e.g 10.73.xx.xx:6789:/
    :param mount_point: mount point, str
    :param options: mount options, e.g -o name=admin,secret=AQCAgd5cFyMQLhAAHaz6w+WKy5LvmKjmxx
    :param verbose: enable verbose or not
    :param session: mount within the session if given
    """
    cephfs_umount(ceph_uri, mount_point, verbose, session)
    mount_cmd = ['mount -t ceph']
    if options:
        mount_cmd.extend(['-o', options])
    mount_cmd.extend([ceph_uri, mount_point])
    mount_cmd = ' '.join(mount_cmd)
    if session:
        create_mount_point = "mkdir -p  %s" % mount_point
        session.cmd(create_mount_point, ok_status=[0, 1], ignore_all_errors=True)
        session.cmd(mount_cmd, ok_status=[0, 1], ignore_all_errors=True)
    else:
        if not os.path.exists(mount_point):
            try:
                utils_misc.make_dirs(mount_point)
            except OSError as dirError:
                logging.debug("Creation of the directory:%s failed:%s", mount_point, str(dirError))
            else:
                logging.debug("Successfully created the directory %s", mount_point)
        process.system(mount_cmd, verbose=verbose)


@error_context.context_aware
def cephfs_umount(ceph_uri, mount_point, verbose=False, session=None):
    """
    Umount ceph volume from mountpoint.

    :param ceph_uri: str to express ceph uri, e.g 10.73.xx.xx:6789:/
    :param mount_point: mount point, str
    :param verbose: enable verbose or not
    :param session: mount within the session if given
    """
    umount_cmd = "umount %s" % mount_point
    if session:
        session.cmd(umount_cmd, ok_status=[0, 1], ignore_all_errors=True)
        rm_mount_point = "rm -rf %s" % mount_point
        session.cmd(rm_mount_point, ok_status=[0, 1], ignore_all_errors=True)
    else:
        process.system(umount_cmd, ignore_status=True, verbose=verbose)
        if os.path.exists(mount_point):
            try:
                utils_misc.safe_rmdir(mount_point)
            except OSError as dirError:
                logging.debug("Delete of the directory:%s failed:%s", mount_point, str(dirError))
            else:
                logging.debug("Successfully deleted the directory %s", mount_point)



"""
Module to control split daemon service.
"""
# pylint: disable=E0611
import re
import logging

import aexpect
from avocado.utils import process
from avocado.utils import wait

from virttest import libvirtd_decorator
from virttest import remote
from virttest import utils_misc
from virttest.staging import service
from virttest.utils_gdb import GDB

IS_MODULAR_DAEMON = {'local': None}


class VirtDaemonCommon(object):

    """
    Common class to manage libvirt split daemon service on host or guest.
    """
    daemon_name = ''
    virsh_cmd = "virsh -c qemu:///system list"

    def __init__(self, daemon_name='', session=None):
        """
        Initialize an service object for virt daemons.

        :param daemon_name: daemon name such as virtqemud, networkd,etc,.
        :param session: An session to guest or remote host.
        """
        if daemon_name:
            self.daemon_name = daemon_name

        self.session = session
        if self.session:
            self.remote_runner = remote.RemoteRunner(session=self.session)
            runner = self.remote_runner.run
        else:
            runner = process.run

        if not self.daemon_name:
            logging.warning("libvirt split daemon service is not available in host, "
                            "utils_daemons module will not function normally")
        self.virtdaemon = service.Factory.create_service(self.daemon_name, run=runner)

    def _wait_for_start(self, timeout=60):
        """
        Wait n seconds for daemon to start. Default is 10 seconds.

        :param timeout: time out for waiting start
        """
        def _check_start():
            try:
                if self.session:
                    self.session.cmd(self.virsh_cmd, timeout=2)
                else:
                    process.run(self.virsh_cmd, timeout=2)
                return True
            except Exception:
                return False
        return utils_misc.wait_for(_check_start, timeout=timeout)

    @libvirtd_decorator.libvirt_version_context_aware_libvirtd_split
    def start(self, reset_failed=True):
        if reset_failed:
            self.virtdaemon.reset_failed()
        if not self.virtdaemon.start():
            return False
        return self._wait_for_start()

    @libvirtd_decorator.libvirt_version_context_aware_libvirtd_split
    def stop(self):
        return self.virtdaemon.stop()

    @libvirtd_decorator.libvirt_version_context_aware_libvirtd_split
    def restart(self, reset_failed=True):
        if reset_failed:
            self.virtdaemon.reset_failed()
        if not self.virtdaemon.restart():
            return False
        return self._wait_for_start()

    @libvirtd_decorator.libvirt_version_context_aware_libvirtd_split
    def is_running(self):
        return self.virtdaemon.status()


class VirtQemud(VirtDaemonCommon):
    """
    Class for manage virtqemud daemon service on host or guest.
    """
    # Override daemon name in super class.
    daemon_name = 'virtqemud'


class VirtProxyd(VirtDaemonCommon):
    """
    Class for manage virtproxyd daemon service on host or guest.
    """
    # Override daemon name in super class.
    daemon_name = 'virtproxyd'


class VirtQemudSession(object):

    """
    Interaction virtqemud daemon session by directly call the virtqemud command.
    With gdb debugging feature can be optionally started.
    """

    def __init__(self, gdb=False,
                 logging_handler=None,
                 logging_params=(),
                 logging_pattern=r'.*'):
        """
        :param gdb: Whether call the session with gdb debugging support
        :param logging_handler: Callback function to handle logging
        :param logging_params: additional logging parameters
        :param logging_pattern: Regex for filtering specific log lines
        """
        self.gdb = None
        self.tail = None
        self.running = False
        self.pid = None
        self.bundle = {"stop-info": None}
        self.virtqemud_service = VirtQemud()
        self.was_running = self.virtqemud_service.is_running()
        if self.was_running:
            logging.debug('Stopping virtqemud service')
            self.virtqemud_service.stop()

        self.logging_handler = logging_handler
        self.logging_params = logging_params
        self.logging_pattern = logging_pattern

        if gdb:
            self.gdb = GDB('virtqemud')
            self.gdb.set_callback('stop', self._stop_callback, self.bundle)
            self.gdb.set_callback('start', self._start_callback, self.bundle)
            self.gdb.set_callback('termination', self._termination_callback)

    def _output_handler(self, line):
        """
        Adapter output callback function.
        """
        if self.logging_handler is not None:
            if re.match(self.logging_pattern, line):
                self.logging_handler(line, *self.logging_params)

    def _termination_handler(self, status):
        """
        Helper aexpect terminaltion handler
        """
        self.running = False
        self.exit_status = status
        self.pid = None

    def _termination_callback(self, gdb, status):
        """
        Termination handler function triggered when virtqemud exited.

        :param gdb: Instance of the gdb session
        :param status: Return code of exited virtqemud session
        """
        self.running = False
        self.exit_status = status
        self.pid = None

    def _stop_callback(self, gdb, info, params):
        """
        Stop handler function triggered when gdb virtqemud stopped.

        :param gdb: Instance of the gdb session
        :param info: information
        :param params: Parameter lists
        """
        self.running = False
        params['stop-info'] = info

    def _start_callback(self, gdb, info, params):
        """
        Stop handler function triggered when gdb virtqemud started.

        :param gdb: Instance of the gdb session
        :param status: Return code of exited virtqemud session
        """
        self.running = True
        params['stop-info'] = None

    def set_callback(self, callback_type, callback_func, callback_params=None):
        """
        Set a customized gdb callback function.
        """
        if self.gdb:
            self.gdb.set_callback(
                callback_type, callback_func, callback_params)
        else:
            logging.error("Only gdb session supports setting callback")

    def start(self, arg_str='', wait_for_working=True):
        """
        Start virtqemud session.

        :param arg_str: Argument passing to the session
        :param wait_for_working: Whether wait for virtqemud finish loading
        """
        if self.gdb:
            self.gdb.run(arg_str=arg_str)
            self.pid = self.gdb.pid
        else:
            self.tail = aexpect.Tail(
                "%s %s" % ('virtqemud', arg_str),
                output_func=self._output_handler,
                termination_func=self._termination_handler,
            )
            self.running = True

        if wait_for_working:
            self.wait_for_working()

    def cont(self):
        """
        Continue a stopped virtqemud session.
        """
        if self.gdb:
            self.gdb.cont()
        else:
            logging.error("Only gdb session supports continue")

    def kill(self):
        """
        Kill the virtqemud session.
        """
        if self.gdb:
            self.gdb.kill()
        else:
            self.tail.kill()

    def restart(self, arg_str='', wait_for_working=True):
        """
        Restart the virtqemud session.

        :param arg_str: Argument passing to the session
        :param wait_for_working: Whether wait for virtqemud finish loading
        """
        logging.debug("Restarting virtqemud session")
        self.kill()
        self.start(arg_str=arg_str, wait_for_working=wait_for_working)

    def wait_for_working(self, timeout=60):
        """
        Wait for virtqemud to work.

        :param timeout: Max wait time
        """
        logging.debug('Waiting for virtqemud to work')
        return utils_misc.wait_for(
            self.is_working,
            timeout=timeout,
        )

    def back_trace(self):
        """
        Get the backtrace from gdb session.
        """
        if self.gdb:
            return self.gdb.back_trace()
        else:
            logging.warning('Can not get back trace without gdb')

    def insert_break(self, break_func):
        """
        Insert a function breakpoint.

        :param break_func: Function at which breakpoint inserted
        """
        if self.gdb:
            return self.gdb.insert_break(break_func)
        else:
            logging.warning('Can not insert breakpoint without gdb')

    def is_working(self):
        """
        Check if virtqemud is start by return status of 'virsh -c qemu:///system list'
        """
        virsh_cmd = "virsh -c qemu:///system list"
        try:
            process.run(virsh_cmd, timeout=2)
            return True
        except process.CmdError:
            return False

    def wait_for_stop(self, timeout=60, step=0.1):
        """
        Wait for virtqemud to stop.

        :param timeout: Max wait time
        :param step: Checking interval
        """
        logging.debug('Waiting for virtqemud to stop')
        if self.gdb:
            return self.gdb.wait_for_stop(timeout=timeout)
        else:
            return wait.wait_for(
                lambda: not self.running,
                timeout=timeout,
                step=step,
            )

    def wait_for_termination(self, timeout=60):
        """
        Wait for virtqemud gdb session to exit.

        :param timeout: Max wait time
        """
        logging.debug('Waiting for virtqemud to terminate')
        if self.gdb:
            return self.gdb.wait_for_termination(timeout=timeout)
        else:
            logging.error("Only gdb session supports wait_for_termination.")

    def exit(self):
        """
        Exit the virtqemud session.
        """
        if self.gdb:
            self.gdb.exit()
        else:
            if self.tail:
                self.tail.close()

        if self.was_running:
            self.virtqemud_service.start()


@libvirtd_decorator.libvirt_version_context_aware_libvirtd_split
def virtqemud_start():
    virtqemud_instance = VirtQemud()
    return virtqemud_instance.start()


@libvirtd_decorator.libvirt_version_context_aware_libvirtd_split
def virtqemud_is_running():
    virtqemud_instance = VirtQemud()
    return virtqemud_instance.is_running()


@libvirtd_decorator.libvirt_version_context_aware_libvirtd_split
def virtqemud_stop():
    virtqemud_instance = VirtQemud()
    return virtqemud_instance.stop()


@libvirtd_decorator.libvirt_version_context_aware_libvirtd_split
def virtqemud_restart():
    virtqemud_instance = VirtQemud()
    return virtqemud_instance.restart()


@libvirtd_decorator.libvirt_version_context_aware_libvirtd_split
def service_virtqemud_control(action, session=None):
    virtqemud_instance = VirtQemud(session)
    getattr(virtqemud_instance, action)()


@libvirtd_decorator.libvirt_version_context_aware_libvirtd_split
def virtproxyd_start():
    virtproxyd_instance = VirtProxyd()
    return virtproxyd_instance.start()


@libvirtd_decorator.libvirt_version_context_aware_libvirtd_split
def virtproxyd_is_running():
    virtproxyd_instance = VirtProxyd()
    return virtproxyd_instance.is_running()


@libvirtd_decorator.libvirt_version_context_aware_libvirtd_split
def virtproxyd_stop():
    virtproxyd_instance = VirtProxyd()
    return virtproxyd_instance.stop()


@libvirtd_decorator.libvirt_version_context_aware_libvirtd_split
def virtproxyd_restart():
    virtproxyd_instance = VirtProxyd()
    return virtproxyd_instance.restart()


def is_modular_daemon(session=None):
    """
    Check whether modular daemon is enabled

    :params session: An session to guest or remote host
    :return: True if modular daemon is enabled
    """
    if session:
        runner = remote.RemoteRunner(session=session).run
        host_key = runner('hostname').stdout_text.strip()
        if host_key not in IS_MODULAR_DAEMON:
            IS_MODULAR_DAEMON[host_key] = None
    else:
        runner = process.run
        host_key = "local"
    if IS_MODULAR_DAEMON[host_key] is None:
        daemons = ["virtqemud.socket", "virtinterfaced.socket",
                   "virtnetworkd.socket", "virtnodedevd.socket",
                   "virtnwfilterd.socket", "virtsecretd.socket",
                   "virtstoraged.socket", "virtproxyd.socket"]

        if any([service.Factory.create_service(d, run=runner).status()
           for d in daemons]):
            IS_MODULAR_DAEMON[host_key] = True
        else:
            IS_MODULAR_DAEMON[host_key] = False
    return IS_MODULAR_DAEMON[host_key]



"""
This module provides interfaces about NVMe storage backend.

Available functions:
- get_image_filename: Get the image filename from NVMe.
- file_exists: Check whether the NVMe image file exists.
- parse_uri: Parse the URI from NVMe image filename.

"""
import re

from avocado.utils import process

from virttest import utils_misc


def get_image_filename(address, namespace):
    """
    Get the image filename from NVMe.

    :param address: The PCI address NVMe,
                    format: $domain:$bus:$slot.$function/$namespace.
    :type address: str
    :param namespace: The namespace number starting according to the NVMe spec.
    :type namespace: str
    :return: The image filename from NVMe,
             the format of nmve://$domain:$bus:$slot.$function/$namespace
             e.g: nvme://0000:44:00.0/1
    :rtype: str
    """
    return 'nvme://%s/%s' % (address, namespace)


def file_exists(params, filename):
    """
    Check whether the NVMe image file exists.

    :param params: A dict containing image parameters.
    :type params: dict
    :param filename: The NVMe image filename.
    :type filename: str
    :return: True if the NVMe image file exists, else False
    :rtype: bool
    """
    cmd = "%s info %s" % (utils_misc.get_qemu_img_binary(params), filename)
    o = process.run(cmd, 60, False, True).stdout_text.strip()
    return params.get('image_format') in o


def parse_uri(filename):
    """
    Get the address and namespace from NVMe image filename.

    :param filename: The NVMe filename,
                     format of nmve://$domain:$bus:$slot.$function/$namespace
    :return: The tuples: (address, namespace)
    :rtype: tuple
    """
    return re.match(r'nvme://(\w+:\w+:\w+\.\w+)/(\w+)', filename).groups()



from threading import Lock
try:
    from collections import UserDict as IterableUserDict
except ImportError:
    from UserDict import IterableUserDict
from collections import OrderedDict

from avocado.core import exceptions

from six.moves import xrange


class ParamNotFound(exceptions.TestSkipError):
    pass


class Params(IterableUserDict):

    """
    A dict-like object passed to every test.
    """
    lock = Lock()

    def __getitem__(self, key):
        """ overrides the error messages of missing params[$key] """
        try:
            return IterableUserDict.__getitem__(self, key)
        except KeyError:
            raise ParamNotFound("Mandatory parameter '%s' is missing. "
                                "Check your cfg files for typos/mistakes" %
                                key)

    def get(self, key, default=None):
        """ overrides the behavior to catch ParamNotFound error"""
        try:
            return self[key]
        except ParamNotFound:
            return default

    def setdefault(self, key, failobj=None):
        if key not in self:
            self[key] = failobj
        return self[key]

    def objects(self, key):
        """
        Return the names of objects defined using a given key.

        :param key: The name of the key whose value lists the objects
                (e.g. 'nics').
        """
        lst = self.get(key, "").split()
        # remove duplicate elements
        objs = list({}.fromkeys(lst).keys())
        # sort list to keep origin order
        objs.sort(key=lst.index)
        del lst
        return objs

    def object_params(self, obj_name):
        """
        Return a dict-like object containing the parameters of an individual
        object.

        This method behaves as follows: the suffix '_' + obj_name is removed
        from all key names that have it.  Other key names are left unchanged.
        The values of keys with the suffix overwrite the values of their
        suffixless versions.

        :param obj_name: The name of the object (objects are listed by the
                objects() method).
        """
        suffix = "_" + obj_name
        self.lock.acquire()
        new_dict = self.copy()
        self.lock.release()
        for key in list(new_dict.keys()):
            if key.endswith(suffix):
                new_key = key.split(suffix)[0]
                new_dict[new_key] = new_dict[key]
        return new_dict

    def object_counts(self, count_key, base_name):
        """
        This is a generator method: to give it the name of a count key and a
        base_name, and it returns an iterator over all the values from params
        """
        count = self.get(count_key, 1)
        # Protect in case original is modified for some reason
        cpy = self.copy()
        for number in xrange(1, int(count) + 1):
            key = "%s%s" % (base_name, number)
            yield (key, cpy.get(key))

    def copy_from_keys(self, keys):
        """
        Return sub dict-like object by keys

        :param keys: white lists of key
        :return: dict-like object
        """
        new_dict = self.copy()
        new_dict.clear()
        for key in keys:
            if self.get(key):
                new_dict[key] = self.get(key)
        return new_dict

    def get_boolean(self, key, default=False):
        """
        Check if a config option is set to a default affirmation or not.

        :param key: config option key
        :type key: str
        :param bool default: whether to assume "yes" or "no" if param
                             does not exist (defaults to False/"no").
        :return whether option is set to 'yes'
        :rtype: bool
        """
        value = self.get(key, "yes" if default else "no")
        if value in ("yes", "on", "true"):
            return True
        if value in ("no", "off", "false"):
            return False
        raise ValueError("Cannot get boolean parameter value for %s: %s", key, value)

    def get_numeric(self, key, default=0, target_type=int):
        """
        Get numeric value converting to integer if necessary.

        :param str key: config option key
        :param int default: default numeric value
        :param type target_type: numeric type to return like int or float
        :return numerical type `target_type` converted parameter value
        :rtype: int or float
        """
        return target_type(self.get(key, default))

    def get_list(self, key, default="", delimiter=None, target_type=str):
        """
        Get a parameter value that is a character delimited list.

        :param str key: parameter whose value is list
        :param str default: default list value
        :param delimiter: character to split list items
        :type delimiter: str or None
        :param type target_type: type of each item, default is string
        :returns: empty list if if the key is not in the parameters
        :rtype: [str]

        .. note:: This an extension to the :py:func:`Params.objects` method as
          it allows for delimiters other than white space.
        .. seealso:: :py:func:`param_dict`
        """
        param_string = self.get(key, default)
        if not param_string:
            return []
        else:
            return [target_type(entry) for entry in param_string.split(delimiter)]

    def get_dict(self, key, default="", delimiter=None, need_order=False):
        """
        Get a param value that has the form 'name1=value1 name2=value2 ...'.

        :param str key: parameter whose value is dict
        :param str default: default dict value
        :param str delimiter: character to split list items
        :type delimiter: str or None
        :param bool need_order: whether to return an OrderedDict instead of
                                a regular dict
        :returns: empty dict if if the key is not in the parameters, a dict
                  or an ordered dictionary if the item order is important
        :rtype: {str: str}

        This uses :py:meth:`get_list` to convert the list entries to dict.
        """
        if need_order:
            result = OrderedDict()
        else:
            result = dict()
        for entry in self.get_list(key, default, delimiter):
            index = entry.find('=')
            if index == -1:
                raise ValueError('failed to find "=" in "{0}" (value for {1})'
                                 .format(entry, key))
            result[entry[:index].strip()] = entry[index+1:].strip()
        return result

    def drop_dict_internals(self):
        """
        Drop internal keys which are not of our concern and
        return the modified params.

        :returns: parameters without the internal keys
        :rtype: {str, str}
        """
        return Params({key: value for key, value in self.items() if not key.startswith("_")})



"""
Class which allows property and dict-like access to a fixed set of instance
attributes.  Attributes are locked by __slots__, however accessor methods
may be created/removed on instances, or defined by the subclass.  An
INITIALIZED attribute is provided to signel completion of __init__()
for use by accessor methods (i.e. so they know when __init__ may be
setting values).

Subclasses must define a __slots__ class attribute containing the list
of attribute names to reserve.  All additional subclass descendents
must explicitly copy __slots__ from the parent in their definition.

Users of subclass instances are expected to get/set/del attributes
only via the standard object or dict-like interface.  i.e.

instance.attribute = whatever
or
instance['attribute'] = whatever

Internally, methods are free to call the accessor methods.  Only
accessor methods should use the special __dict_*__() and __super_*__() methods.
These are there to allow convenient access to the internal dictionary
values and subclass-defined attributes (such as __slots__).

example:

::

    class A(PropCan):
       # Class with *attributes*
        __slots__ = ('a', 'b')
        # 'a' has defined a set/get/del by definition of method with prefix
        #     set_a, get_a, del_a
        # 'b' doesn't have defined set/get/del then classic set/get/del will be
        #     called instead.


        def __init__(self, a=1, b='b'):
           super(A, self).__init__(a, b)


        def set_a(self, value)
            # If is_instance(obj, A) then obj.a = "val" call this method.
            self.__dict_set__("a", value)


        def get_a(self, value)
            # If is_instance(obj, A) then xx = obj.a call this method.
            return self.__dict_get__("a")


        def del_a(self, value)
            # If is_instance(obj, A) then del obj.a call this method.
            self.__dict_del__("a")


    class B(PropCan):
       # Class without *attributes*
       # ***** Even if class doesn't have attributes there should be
       # defined __slots__ = []. Because it is preferred by new style of class.
       # *****
        __slots__ = []


        def __init__(self):
           super(B, self).__init__()
"""


class PropCanInternal(object):

    """
    Semi-private methods for use only by PropCanBase subclasses (NOT instances)
    """
    #: Store all slots for debugging purposes
    __all_slots__ = None

    # The following methods are intended for use by accessor-methods
    # where they may need to bypass the special attribute/key handling

    def __dict_get__(self, key):
        """
        Get a key unconditionally, w/o checking for accessor method or __slots__
        """
        return dict.__getitem__(self, key)

    def __dict_set__(self, key, value):
        """
        Set a key unconditionally, w/o checking for accessor method or __slots__
        """
        dict.__setitem__(self, key, value)

    def __dict_del__(self, key):
        """
        Del key unconditionally, w/o checking for accessor method or __slots__
        """
        return dict.__delitem__(self, key)

    def __super_get__(self, key):
        """
        Get attribute unconditionally, w/o checking accessor method or __slots__
        """
        return object.__getattribute__(self, key)

    def __super_set__(self, key, value):
        """
        Set attribute unconditionally, w/o checking accessor method or __slots__
        """
        object.__setattr__(self, key, value)

    def __super_del__(self, key):
        """
        Del attribute unconditionally, w/o checking accessor method or __slots__
        """
        object.__delattr__(self, key)


class classproperty(property):

    def __get__(self, obj, type_):
        data = self.fget.__get__(None, type_)()
        return data

    def __set__(self, obj, value):
        cls = type(obj)
        return self.fset.__get__(None, cls)(value)


class PropCanBase(dict, PropCanInternal):

    """
    Objects with optional accessor methods and dict-like access to fixed set of keys
    """

    # get_*(), set_*(), del_*() accessor methods called from subclass
    # __init__ sometimes need special handling, this is the signal.
    INITIALIZED = False

    # Help debugging by making all slot values available in all subclasses
    # cache the value on first call
    ___all_slots__ = None

    @classproperty
    @classmethod
    def __all_slots__(cls):
        if not cls.___all_slots__:
            all_slots = []
            for cls_slots in [getattr(_cls, '__slots__', [])
                              for _cls in cls.__mro__]:
                all_slots += cls_slots
            cls.___all_slots__ = tuple(all_slots)
        return cls.___all_slots__

    def __new__(cls, *args, **dargs):
        if not hasattr(cls, '__slots__'):
            raise NotImplementedError("Class '%s' must define __slots__ "
                                      "property" % str(cls))
        newone = super(PropCanBase, cls).__new__(cls, *args, **dargs)
        cls.___all_slots__ = tuple()
        return newone

    def __init__(self, *args, **dargs):
        """
        Initialize contents directly or by way of accessors

        :param args: Initial values for __slots__ keys, same as dict.
        :param dargs: Initial values for __slots__ keys, same as dict.
        """
        # Params are initialized here, not in super
        super(PropCanBase, self).__init__()
        # No need to re-invent dict argument processing
        values = dict(*args, **dargs)
        for key in self.__all_slots__:
            value = values.get(key, "@!@!@!SENTINEL!@!@!@")
            if value != "@!@!@!SENTINEL!@!@!@":
                # Call accessor methods if present
                self[key] = value
        # Let accessor methods know initialization is complete
        self.__super_set__('INITIALIZED', True)

    def __getitem__(self, key):
        try:
            accessor = super(PropCanBase,
                             self).__getattribute__('get_%s' % key)
        except AttributeError:
            return super(PropCanBase, self).__getitem__(key)
        return accessor()

    def __setitem__(self, key, value):
        self.__canhaz__(key, KeyError)
        try:
            accessor = super(PropCanBase,
                             self).__getattribute__('set_%s' % key)
        except AttributeError:
            return super(PropCanBase, self).__setitem__(key, value)
        return accessor(value)

    def __delitem__(self, key):
        try:
            accessor = super(PropCanBase,
                             self).__getattribute__('del_%s' % key)
        except AttributeError:
            return super(PropCanBase, self).__delitem__(key)
        return accessor()

    def __get__(self, key, objtype=None):
        try:
            # Attempt to call accessor methods first whenever possible
            self.__canhaz__(key, KeyError)
            return self.__getitem__(key)
        except KeyError:
            # Allow subclasses to define attributes if required
            return super(PropCanBase, self).__getattribute__(key)

    def __set__(self, key, value):
        self.__canhaz__(key)
        try:
            return self.__setitem__(key, value)
        except KeyError as detail:
            # Prevent subclass instances from defining normal attributes
            raise AttributeError(str(detail))

    def __getattr__(self, key):
        try:
            # Attempt to call accessor methods first whenever possible
            self.__canhaz__(key, KeyError)
            return self.__getitem__(key)
        except KeyError:
            # Allow subclasses to define attributes if required
            return super(PropCanBase, self).__getattribute__(key)

    def __setattr__(self, key, value):
        self.__canhaz__(key)
        try:
            return self.__setitem__(key, value)
        except KeyError as detail:
            # Prevent subclass instances from defining normal attributes
            raise AttributeError(str(detail))

    def __delattr__(self, key):
        self.__canhaz__(key)
        try:
            return self.__delitem__(key)
        except KeyError as detail:
            # Prevent subclass instances from deleting normal attributes
            raise AttributeError(str(detail))

    def __canhaz__(self, key, excpt=AttributeError):
        """
        Quickly determine if an accessor or instance attribute name is defined.
        """
        slots = self.__all_slots__
        keys = slots + ('get_%s' % key, 'set_%s' % key, 'del_%s' % key)
        if key not in keys:
            raise excpt("Key '%s' not found in super class attributes or in %s"
                        % (str(key), str(keys)))

    def copy(self):
        """
        Copy properties by value, not by reference.
        """
        return self.__class__(dict(self))

    def update(self, other=None, excpt=AttributeError, **kwargs):
        """
        Update properties in __all_slots__ with another dict.
        """
        _tmp_dict = dict()
        _other_dict = dict()

        if other:
            _other_dict = dict(other)

        try:
            _tmp_dict.update(_other_dict)
            _tmp_dict.update(kwargs)
        except TypeError as detail:
            raise excpt(detail)

        for item in list(_tmp_dict.keys()):
            self[item] = _tmp_dict[item]


class PropCan(PropCanBase):

    """
    Special value handling on retrieval of None values
    """

    def __len__(self):
        length = 0
        for key in self.__all_slots__:
            # special None/False value handling
            if self.__contains__(key):
                length += 1
        return length

    def __contains__(self, key):
        try:
            value = self.__dict_get__(key)
        except (KeyError, AttributeError):
            return False
        # Avoid inf. recursion if value == self
        if issubclass(type(value), type(self)) or value:
            return True
        return False

    def __eq__(self, other):
        # special None/False value handling
        return dict([(key, value) for key, value in list(self.items())]) == other

    def __ne__(self, other):
        return not self.__eq__(other)

    def keys(self):
        # special None/False value handling
        return [key for key in self.__all_slots__
                if self.__contains__(key)]

    def values(self):
        # special None/False value handling
        return [self[key] for key in list(self.keys())]

    def items(self):
        return tuple([(key, self[key]) for key in list(self.keys())])

    has_key = __contains__

    def set_if_none(self, key, value):
        """
        Set the value of key, only if it's not set or None
        """
        if key not in self:
            self[key] = value

    def set_if_value_not_none(self, key, value):
        """
        Set the value of key, only if value is not None
        """
        if value:
            self[key] = value



"""
A module is used to interact with VMware Vshpere Server

For more examples of pyvmomi, please refer:
https://github.com/vmware/pyvmomi-community-samples
"""
import datetime
import logging

from functools import wraps

from pyVim.connect import SmartConnect, SmartConnectNoSSL, Disconnect
from pyVim.task import WaitForTask
from pyVmomi import vim


def to_list(obj):
    tmp_list = []

    if not obj:
        return tmp_list

    if isinstance(obj, list):
        tmp_list.extend(obj)
    else:
        tmp_list.append(obj)
    return tmp_list


class VSphereError(Exception):
    """
    A common error for this module
    """

    def __init__(self, msg):
        self.msg = msg


class VSphereVMNotSpecified(VSphereError):
    """
    An error when the VM was not set but called vm related operation
    """

    def __init__(self, msg=None):
        if not msg:
            msg = "VM not specified"
        super(VSphereVMNotSpecified, self).__init__(msg)
        self.msg = msg

    def __str__(self):
        return self.msg


class VSphereVMNotFound(VSphereError):
    """
    An error when the VM was not found
    """

    def __init__(self, vm_name, msg=None):
        if not msg:
            msg = "VM not Found"
        super(VSphereVMNotFound, self).__init__(msg)
        self.vm = vm_name
        self.msg = msg

    def __str__(self):
        msg = '%s: %s' % (self.msg, self.vm)
        return msg


class VSphereDevNoChangeId(VSphereError):
    """
    An error when the VM was not found
    """

    def __init__(self, vm_name, dev, msg=None):
        if not msg:
            msg = "ChangeID not Found"
        super(VSphereDevNoChangeId, self).__init__(msg)
        self.vm = vm_name
        self.dev = dev
        self.msg = msg

    def __str__(self):
        msg = "%s for VM(%s): device(label='%s' key=%s)" % (
            self.msg, self.vm, self.dev.deviceInfo.label, self.dev.key)
        return msg


class VSphereSnapNotFound(VSphereError):
    """
    An error when the snapshot of the VM was not found
    """

    def __init__(self, vm_name, snapshot_id, msg=None):
        if not msg:
            msg = "snapshot not Found"
        super(VSphereSnapNotFound, self).__init__(msg)
        self.vm = vm_name
        self.snapshot_id = snapshot_id
        self.msg = msg

    def __str__(self):
        msg = '%s for VM %s: %s' % (self.msg, self.vm, self.snapshot_id)
        return msg


class VSphereInvalidDevType(VSphereError):
    """
    An error when the snapshot of the VM was not found
    """

    def __init__(self, dev_type, msg=None):
        if not msg:
            msg = "Invalid device Type"
        super(VSphereInvalidDevType, self).__init__(msg)
        self.dev_type = dev_type
        self.msg = msg

    def __str__(self):
        msg = '%s: %s' % (self.msg, self.dev_type)
        return msg


class VSphereConnection(object):
    """
    This is a useful class when you want to do some temporary
    operation. The class can do initialization and cleanup
    automatically.
    Besides the necessary parameters of VSphere, you can pass
    a 'vm_name' parameter to set the target_vm automatically.

    Examples:
    >>> from utils_pyvmomi import *
    >>> connect_args = {'host': "x.x.x.x", 'user': 'root', 'pwd': 'xxx', 'vm_name': 'pyvmomi-rhel8.1'}
    >>> with VSphereConnection(**connect_args) as conn:
        ...     print(conn.get_vm_summary()['mac_address'])
        ...
        ['00:50:56:ac:09:3e']

    or set the target_vm later.

    >>> connect_args = {'host': "x.x.x.x", 'user': 'root', 'pwd': 'xxx'}
    >>> with VSphereConnection(**connect_args) as conn:
    ...     conn.target_vm = 'pyvmomi-rhel8.1'
    ...     print(conn.get_vm_summary()['mac_address'])
    ...
    ['00:50:56:ac:09:3e']

    """

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self.vsphere_conn = None

        try:
            self.vm_name = self.kwargs.pop('vm_name')
        except KeyError:
            self.vm_name = None

    def __enter__(self):
        self.vsphere_conn = VSphere(*self.args, **self.kwargs)
        self.vsphere_conn.connect()
        if self.vm_name:
            self.vsphere_conn.target_vm = self.vm_name
        return self.vsphere_conn

    def __exit__(self, *exc_info):
        self.vsphere_conn.close()


class VSphere(object):
    """
    The VShpere class is used to do interactive with vshpere server.

    It's a convient class for testing, Users don't have to spend much
    time on pyVmomi.

    By this class, you can query basic information of a VM, do power
    on/off, create snapshot, etc.

    More functions will be added if it's necessary for testing.

    Examples:

    >>> connect_args = {'host': "x.x.x.x", 'user': 'root', 'pwd': 'xxx'}
    >>> conn = VSphere(**connect_args)
    >>> conn.target_vm = 'pyvmomi-rhel8.1'
    >>> print(conn.get_vm_summary()['mac_address'])
    ['00:50:56:ac:09:3e']
    >>> conn.close()
    """

    def __init__(self, insecure=True, **kwargs):
        """
        An initialization function for VShpere Class

        For the full supported parameters, please refer:
        https://github.com/vmware/pyvmomi/blob/master/pyVim/connect.py

        Some basic parameter are:

        :param host: ip or hostname of vshpere server
        :param port: a port number, default is 443
        :param user: user name, default is 'root'
        :param pwd: password
        :param insecure: If true, the connection will have no SSL.
                         Default is True.
        """
        self.insecure = insecure
        self.kwargs = kwargs

        # An service instance to VSphere server
        self.service_instance = None

        # An internal vm object.
        # Users should use 'target_vm' instead of this one.
        self._target_vm = None
        # Use to recover the conn when it's dead
        self._target_vm_name = None

    def connect(self):
        """
        Initialize the service instance to the VSphere server
        """
        # Check if an valid connection has already been establised.
        # If yes, just refresh the connection to keep it alive.
        # If not, close the old connection and establise a new one.
        if self.is_conn_dead():
            self.close()
        else:
            self.keep_alive()
            return

        kwargs = self.kwargs

        if self.insecure:
            self.service_instance = SmartConnectNoSSL(**kwargs)
        else:
            self.service_instance = SmartConnect(**kwargs)

        if self.service_instance:
            logging.debug(
                'New vsphere connection established: %s (%s)',
                self.service_instance, id(self.service_instance))

    def close(self):
        """
        Cleanup the vsphere resources
        """
        del self.target_vm
        if not self.service_instance:
            return
        logging.debug('vsphere connection closed: %s (%s)',
                      self.service_instance, id(self.service_instance))
        Disconnect(self.service_instance)
        self.service_instance = None

    def keep_alive(self):
        """
        Get the time of vshpere server to keep the connection alive
        """
        self.service_instance.serverClock

    def is_conn_dead(self):
        """
        Check if the connection is dead.

        If it's dead, return True, else
        return False.
        """
        if not self.service_instance:
            return True
        try:
            self.keep_alive()
        except vim.fault.NotAuthenticated:
            return True
        return False

    # pylint: disable=E0213
    def vm_picker(f):
        """
        A decoractor function is used to determine if which VM object
        should be used.

        Users can specifiy an vshpere vm object to 'vm_obj', or a vm
        name to 'vm_name'. This function can help to find the vm by its
        name.

        If both 'vm_obj' and 'vm_name' are passed, 'vm' will be used instead
        of 'vm_name'.

        Note: If target_vm was set, then you do get_mac_address(vm_name='another-vm'),
        It will not work as expected. Because the target_vm will be used
        instead of new vm specified by 'vm_name'. You need to change
        target_vm by 'self.target_vm = new_vm_name'

        Examples:

        >>> conn.target_vm.name
        'pyvmomi-rhel8.1'
        >>> conn.get_mac_address(vm_obj=conn.target_vm)
        ['00:50:56:ac:09:3e']
        >>> conn.get_mac_address(vm_name='esx6.5-rhel6.9-i386')
        ['00:50:56:ac:09:3e']
        >>> conn.target_vm = 'esx6.5-rhel6.9-i386'
        >>> conn.get_mac_address(vm_name='esx6.5-rhel6.9-i386')
        ['00:50:56:a2:ee:5e']

        """
        @wraps(f)
        def wraper(self, *args, **kwargs):
            vmobj = kwargs.get('vm_obj')
            vm_name = kwargs.get('vm_name')
            if not vmobj:
                if self.target_vm:
                    # If conn is dead, reassign the
                    # vm to active a new conn.
                    if self.is_conn_dead():
                        self.target_vm = self._target_vm_name
                    vmobj = self.target_vm
                    if vm_name:
                        logging.warning(
                            "Have you forgotten to reset target_vm to 'new vm name'?")
                elif vm_name:
                    self.target_vm = vm_name
                    vmobj = self.target_vm
                else:
                    raise VSphereVMNotSpecified

                kwargs.update({'vm_obj': vmobj})

            # pylint: disable=E1102
            return f(self, *args, **kwargs)
        return wraper

    def get_all_vms(self):
        """
        Return all VMs on VSphere
        """
        self.connect()
        content = self.service_instance.RetrieveContent()
        container_view = content.viewManager.CreateContainerView(
            self.service_instance.RetrieveContent().rootFolder, [vim.VirtualMachine], True)
        return [vm for vm in container_view.view]

    def _get_vm(self):
        """
        Get VM
        """
        return self._target_vm

    def _set_vm(self, name):
        """
        Set VM by name

        :param name: a vm's name
        """
        # For vm recovery if faied
        tmp_vm = self._target_vm
        # Clean up old VM
        self._target_vm = None
        # Connect to VSphere server
        self.connect()
        # Could two VM names be same on VSphere?
        # Right now the code only returns the first found VM
        for vm in self.get_all_vms():
            if vm.name == name:
                self._target_vm = vm
                break
        # Not Found
        if not self._target_vm:
            # Restore the target_vm to old value
            self._target_vm = tmp_vm
            raise VSphereVMNotFound(name)
        self._target_vm_name = self._target_vm.name
        logging.debug('Current target VM is %s' % self._target_vm.name)

    def _del_vm(self):
        """
        Clean up the vm but not delete it
        """
        self._target_vm = None

    # This property stands a target vm object in vsphere.
    # e.g. vim.VirtualMachine:vm-1464
    target_vm = property(
        _get_vm,
        _set_vm,
        _del_vm,
        "The target VM to check and update.")

    @vm_picker
    def get_mac_address(self, vm_obj=None, vm_name=None):
        """
        Return all mac addresses of the VM

        :param vm_obj: a vsphere vm object
        :param vm_name: a vm's name
        """
        mac_list = []
        for dev in vm_obj.config.hardware.device:
            if isinstance(dev, vim.vm.device.VirtualEthernetCard):
                mac_list.append(dev.macAddress)
        return mac_list

    @vm_picker
    def get_vm_summary(self, vm_obj=None, vm_name=None):
        """
        Return some configurations of the VM

        Note: For 'ip_address', It takes some time for the guest
        to obtain the value. You need to keep trying it.

        :param vm_obj: a vsphere vm object
        :param vm_name: a vm's name
        """
        vm_summary = {}
        raw_vmsummary = vm_obj.summary

        # vim.vm.Summary.ConfigSummary
        vm_summary['name'] = raw_vmsummary.config.name
        vm_summary['memory_mb'] = raw_vmsummary.config.memorySizeMB
        vm_summary['vm_path'] = raw_vmsummary.config.vmPathName
        vm_summary['num_cpu'] = raw_vmsummary.config.numCpu
        vm_summary['num_cores_per_socket'] = vm_obj.config.hardware.numCoresPerSocket
        vm_summary['num_ethernet_cards'] = raw_vmsummary.config.numEthernetCards
        vm_summary['num_virtual_disks'] = raw_vmsummary.config.numVirtualDisks
        vm_summary['uuid'] = raw_vmsummary.config.uuid
        vm_summary['instance_uuid'] = raw_vmsummary.config.instanceUuid

        # vim.vm.summary.runtime
        vm_summary['power_state'] = raw_vmsummary.runtime.powerState

        # vim.vm.summary.guest
        vm_summary['hostname'] = raw_vmsummary.guest.hostName
        vm_summary['ip_address'] = raw_vmsummary.guest.ipAddress
        vm_summary['mac_address'] = self.get_mac_address(
            vm_obj=vm_obj, vm_name=vm_name)

        return vm_summary

    @vm_picker
    def power_on(self, vm_obj=None, vm_name=None):
        """
        Power on the VM

        :param vm_obj: a vsphere vm object
        :param vm_name: a vm's name
        """
        WaitForTask(vm_obj.PowerOn())
        logging.debug('VM %s was powered on', vm_obj.name)

    @vm_picker
    def power_off(self, vm_obj=None, vm_name=None):
        """
        Force power off the VM

        This may cause windows guests to enter emergency
        recovery mode.

        :param vm_obj: a vsphere vm object
        :param vm_name: a vm's name
        """
        WaitForTask(vm_obj.PowerOff())
        logging.debug('VM %s was powered off', vm_obj.name)

    @vm_picker
    def remove_all_snapshots(self, vm_obj=None, vm_name=None):
        """
        Remove all snapshots of the VM

        :param vm_obj: a vsphere vm object
        :param vm_name: a vm's name
        """
        if not vm_obj.snapshot:
            return
        logging.debug('Remove all snapshots for VM %s', vm_obj.name)
        WaitForTask(vm_obj.RemoveAllSnapshots())

    @vm_picker
    def remove_current_snapshot(self, vm_obj=None, vm_name=None):
        """
        Remove current snapshot of the VM

        :param vm_obj: a vsphere vm object
        :param vm_name: a vm's name
        """
        if not vm_obj.snapshot:
            return
        logging.debug('Remove current snapshot for VM %s', vm_obj.name)
        WaitForTask(
            vm_obj.snapshot.currentSnapshot.Remove(
                removeChildren=True))

    @vm_picker
    def find_snapshot_by_id(self, snapshot_id, vm_obj=None, vm_name=None):
        """
        Find a vm's snapshot by the snapshot ID

        The snapshot ID should be saved when creating a snapshot if you
        need to use it in this function.

        A snapshot obj in pyvmomi is like 'vim.vm.Snapshot:snapshot-1474',
        the snapshot ID is 'snapshot-1474'

        :param snapshot_id: a snapshot ID
        :param vm_obj: a vsphere vm object
        :param vm_name: a vm's name
        """
        def _find_snapshot_by_recursive(snap, snapshot_id):
            if snapshot_id == snap.snapshot._moId:
                return snap.snapshot
            for child_snap in snap.childSnapshotList:
                return _find_snapshot_by_recursive(child_snap, snapshot_id)
            return None

        if not vm_obj.snapshot:
            return None

        for snap_tree in vm_obj.snapshot.rootSnapshotList:
            return _find_snapshot_by_recursive(snap_tree, snapshot_id)
        # This return should not be executed in theory.
        return None

    @vm_picker
    def remove_snapshot_by_id(
            self,
            snapshot_id,
            vm_obj=None,
            vm_name=None,
            remove_children=True,
            raise_not_found=False):
        """
        Remove a vm's snapshot by the snapshot ID

        The snapshot ID should be saved when creating a snapshot if you
        need to use it in this function.

        A snapshot obj in pyvmomi is like 'vim.vm.Snapshot:snapshot-1474',
        the snapshot ID is 'snapshot-1474'

        :param snapshot_id: a snapshot ID
        :param vm_obj: a vsphere vm object
        :param vm_name: a vm's name
        :param remove_children: whether to remove snapshot's children snapshots
        :param raise_not_found: whether to raise exception if snapshot_id not found
        """
        snap = self.find_snapshot_by_id(snapshot_id)
        if not snap:
            if raise_not_found:
                raise VSphereSnapNotFound(vm_obj.name, snapshot_id)
            else:
                logging.debug(
                    'Not found snapshot_id %s for VM %s',
                    snapshot_id,
                    vm_obj.name)
                return

        logging.debug('Remove snapshot %s for VM %s', snap, vm_obj.name)
        WaitForTask(snap.Remove(removeChildren=remove_children))

    @vm_picker
    def create_snapshot(self, vm_obj=None, vm_name=None):
        """
        Create a snapshot for the VM

        The snapshot ID should be saved when creating a snapshot in order
        to use it later.

        :param vm_obj: a vsphere vm object
        :param vm_name: a vm's name
        """
        # Enable CBT for change tracking
        config_spec = vim.vm.ConfigSpec(changeTrackingEnabled=True)
        WaitForTask(vm_obj.Reconfigure(config_spec))
        WaitForTask(
            vm_obj.CreateSnapshot(
                name='Testing Snapshot CBT',
                description='Created at %s by Avocado-VT on testing host' %
                str(
                    datetime.datetime.now()),
                memory=False,
                quiesce=True))

    @vm_picker
    def get_hardware_devices(
            self,
            vm_obj=None,
            vm_name=None,
            devices=None,
            dev_type=None):
        """
        Return all hardware devices of the VM or filter the devices
        by dev_type.

        If not specified a device type, all devices will be returned.
        If some devices are passed, this function can filter those devices by dev_type.

        Possible values for 'dev_type' are as below:
            vim.vm.device.VirtualCdrom, vim.vm.device.VirtualController,
            vim.vm.device.VirtualDisk, vim.vm.device.VirtualEthernetCard,
            vim.vm.device.VirtualFloppy, vim.vm.device.VirtualKeyboard,
            vim.vm.device.VirtualMachineVideoCard, vim.vm.device.VirtualMachineVMCIDevice,
            vim.vm.device.VirtualMachineVMIROM, vim.vm.device.VirtualNVDIMM,
            vim.vm.device.VirtualParallelPort, vim.vm.device.VirtualPCIPassthrough,
            vim.vm.device.VirtualPointingDevice, vim.vm.device.VirtualSCSIPassthrough,
            vim.vm.device.VirtualSerialPort, vim.vm.device.VirtualSoundCard,
            vim.vm.device.VirtualTPM, vim.vm.device.VirtualUSB

        :param vm_obj: a vsphere vm object
        :param vm_name: a vm's name
        :param devices: a device object return by pyvmomi
        :param dev_type: the type of the device.
        """
        vm_devs = devices if devices else vm_obj.config.hardware.device

        if not dev_type:
            return vm_devs

        if dev_type and issubclass(dev_type, vim.vm.device.VirtualDevice):
            devs = []
            for dev in vm_devs:
                if isinstance(dev, dev_type):
                    devs.append(dev)
            return devs
        else:
            raise VSphereInvalidDevType(dev_type)

    def get_dev_by_key_or_label(self, devices, label=None, key=None):
        """
        Get devices by key or label.

        Return device with label matched or key matched.
        If both label and key are specified, the device matches both
        of them is returned.

        :param devices: a device object return by pyvmomi
        :param label: a device's label
        :param key: A unique value to identify a device
        """

        devs = to_list(devices)

        conds = []
        if label:
            conds.append(lambda dev: dev.deviceInfo.label == label)
        if key:
            conds.append(lambda dev: dev.key == key)
        if not conds:
            raise VSphereError("At least a label or a key must be specified")

        res = [dev for dev in devs if all([cond(dev) for cond in conds])]
        if len(res) > 1:
            raise VSphereError(
                "Mutiple devices are found for label(%s) or key(%s)" %
                (label, key))

        if not res:
            raise VSphereError(
                "Not found device for label(%s) or key(%s)" %
                (label, key))

        logging.debug(
            "Found device: label(%s) key(%s) summary(%s)",
            res[0].deviceInfo.label,
            res[0].key,
            res[0].deviceInfo.summary)
        return res[0]

    @vm_picker
    def query_changed_disk_areas(
            self,
            vm_obj=None,
            vm_name=None,
            start_offset=0,
            disk_label=None,
            disk_key=None,
            snapshot=None,
            snapshot_id=None):
        """
        Get a list of areas of a virtual disk belonging to this VM that
        have been modified since a well-defined point in the past.
        The beginning of the change interval is identified by "changeId",
        while the end of the change interval is implied by the snapshot ID
        passed in.

        :param vm_obj: a vsphere vm object
        :param vm_name: a vm's name
        :param disk_label: a device's label
        :param disk_key: A unique value to identify a device
        :param start_offset: Start Offset in bytes at which to start
            computing changes. Typically, callers will make multiple
            calls to this function, starting with startOffset 0 and
            then examine the "length" property in the returned
            DiskChangeInfo structure, repeatedly calling queryChangedDiskAreas
            until a map forthe entire virtual disk has been obtained.
        :param snapshot: Snapshot for which changes that have been made
            sine "changeId" should be computed. If not set, changes are
            computed against the "current" snapshot of the virtual machine.
            However, using the "current" snapshot will only work for
            virtual machines that are powered off.
        :param snapshot_id: a snapshot ID
        """
        # If not set snapshot but snapshot_id, find it by
        # snapshot id.
        if not snapshot and snapshot_id:
            snapshot = self.find_snapshot_by_id(snapshot_id)

        if not snapshot:
            all_devs = vm_obj.snapshot.currentSnapshot.config.hardware.device
        else:
            all_devs = snapshot.config.hardware.device

        disk_devs = self.get_hardware_devices(
            devices=all_devs, dev_type=vim.vm.device.VirtualDisk)
        dev = self.get_dev_by_key_or_label(disk_devs, disk_label, disk_key)
        if not dev.backing.changeId:
            raise VSphereDevNoChangeId(vm_obj.name, dev)

        # if snapshot is None, currentsnapshot is used
        return vm_obj.QueryChangedDiskAreas(
            snapshot, dev.key, start_offset, dev.backing.changeId)



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2015
# Author: Lucas Meneghel Rodrigues <lmr@redhat.com>

"""
Avocado VT plugin
"""

import imp
import logging
import os
import sys
import pickle
import pipes
import traceback

from avocado.core import exceptions
from avocado.core import test
from avocado.utils import stacktrace
from avocado.utils import process
from avocado.utils import genio

from virttest import asset
from virttest import error_event
from virttest import bootstrap
from virttest import data_dir
from virttest import env_process
from virttest import funcatexit
from virttest import utils_env
from virttest import utils_params
from virttest import utils_misc
from virttest import version

# avocado-vt no longer needs autotest for the majority of its functionality,
# except by:
# 1) Run autotest on VMs
# 2) Multi host migration
# 3) Proper avocado-vt test status handling
# As in those cases we might want to use autotest, let's have a way for
# users to specify their autotest from a git clone location.
AUTOTEST_PATH = None

if 'AUTOTEST_PATH' in os.environ:
    AUTOTEST_PATH = os.path.expanduser(os.environ['AUTOTEST_PATH'])
    CLIENT_DIR = os.path.join(os.path.abspath(AUTOTEST_PATH), 'client')
    SETUP_MODULES_PATH = os.path.join(CLIENT_DIR, 'setup_modules.py')
    if not os.path.exists(SETUP_MODULES_PATH):
        raise EnvironmentError("Although AUTOTEST_PATH has been declared, "
                               "%s missing." % SETUP_MODULES_PATH)
    SETUP_MODULES = imp.load_source('autotest_setup_modules',
                                    SETUP_MODULES_PATH)
    SETUP_MODULES.setup(base_path=CLIENT_DIR,
                        root_module_name="autotest.client")


BG_ERR_FILE = "background-error.log"


def cleanup_env(env_filename, env_version):
    """
    Pickable function to initialize and destroy the virttest env
    """
    env = utils_env.Env(env_filename, env_version)
    env.destroy()


class VirtTest(test.Test):

    """
    Minimal test class used to run a virt test.
    """

    env_version = utils_env.get_env_version()

    def __init__(self, methodName='runTest', name=None, params=None,
                 base_logdir=None, job=None, runner_queue=None,
                 vt_params=None):
        """
        :note: methodName, name, base_logdir, job and runner_queue params
               are inherited from test.Test
        :param params: avocado/multiplexer params stored as
                       `self.avocado_params`.
        :param vt_params: avocado-vt/cartesian_config params stored as
                          `self.params`.
        """
        self.__params_vt = None
        self.__avocado_params = None
        self.bindir = data_dir.get_root_dir()
        self.virtdir = os.path.join(self.bindir, 'shared')
        # self.__params_vt must be initialized after super
        params_vt = utils_params.Params(vt_params)
        # for timeout use Avocado-vt timeout as default but allow
        # overriding from Avocado params (varianter)
        self.timeout = params_vt.get("test_timeout", self.timeout)

        self.iteration = 0
        self.resultsdir = None
        self.background_errors = error_event.error_events_bus
        # clear existing error events
        self.background_errors.clear()
        super(VirtTest, self).__init__(methodName=methodName, name=name,
                                       params=params,
                                       base_logdir=base_logdir, job=job,
                                       runner_queue=runner_queue)
        self.builddir = os.path.join(self.workdir, 'backends',
                                     vt_params.get("vm_type", ""))
        self.tmpdir = os.path.dirname(self.workdir)
        # Move self.params to self.avocado_params and initialize virttest
        # (cartesian_config) params
        try:
            self.__avocado_params = super(VirtTest, self).params
        except AttributeError:
            # 36LTS set's `self.params` instead of having it as a property
            # which stores the avocado params in `self.__params`
            self.__avocado_params = self.__params
        self.__params_vt = params_vt
        self.debugdir = self.logdir
        self.resultsdir = self.logdir
        utils_misc.set_log_file_dir(self.logdir)
        self.__status = None
        self.__exc_info = None

    @property
    def params(self):
        """
        Avocado-vt test params

        During `avocado.Test.__init__` this reports the original params but
        once the Avocado-vt params are set it reports those instead. This
        is necessary to complete the `avocado.Test.__init__` phase
        """
        if self.__params_vt is not None:
            return self.__params_vt
        else:
            # The `self.__params` is set after the `avocado.test.__init__`,
            # but in newer Avocado `self.params` is used during `__init__`
            # Report the parent's value in such case.
            return super(VirtTest, self).params

    @params.setter
    def params(self, value):
        """
        For compatibility with 36lts we need to support setter on params
        """
        self.__params_vt = value

    @property
    def avocado_params(self):
        """
        Original Avocado (multiplexer/varianter) params
        """
        return self.__avocado_params

    @property
    def datadir(self):
        """
        Returns the path to the directory that contains test data files

        For VT tests, this always returns None. The reason is that
        individual VT tests do not map 1:1 to a file and do not provide
        the concept of a datadir.
        """
        return None

    @property
    def filename(self):
        """
        Returns the name of the file (path) that holds the current test

        For VT tests, this always returns None. The reason is that
        individual VT tests do not map 1:1 to a file.
        """
        return None

    def get_state(self):
        """
        Pre Avocado-60.0 used to override self.__params attribute and
        requires special handling while reporting the state.

        TODO: Remove when 52LTS is deprecated.
        """
        state = super(VirtTest, self).get_state()
        if state["params"] == self.__params_vt:
            state["params"] = self.avocado_params
        return state

    def write_test_keyval(self, d):
        self.whiteboard = str(d)

    def verify_background_errors(self):
        """
        Verify if there are any errors that happened on background threads.
        Logs all errors in the background_errors into background-error.log and
        error the test.
        """
        err_file_path = os.path.join(self.logdir, BG_ERR_FILE)
        bg_errors = self.background_errors.get_all()
        error_messages = ["BACKGROUND ERROR LIST:"]
        for index, error in enumerate(bg_errors):
            error_messages.append(
                "- ERROR #%d -\n%s" % (index, "".join(
                    traceback.format_exception(*error)
                    )))
        genio.write_file(err_file_path, '\n'.join(error_messages))
        if bg_errors:
            msg = ["Background error"]
            msg.append("s are" if len(bg_errors) > 1 else " is")
            msg.append((" detected, please refer to file: "
                        "'%s' for more details.") % BG_ERR_FILE)
            self.error(''.join(msg))

    def __safe_env_save(self, env):
        """
        Treat "env.save()" exception as warnings

        :param env: The virttest env object
        :return: True on failure
        """
        try:
            env.save()
        except Exception as details:
            try:
                pickle.dumps(env.data)
            except Exception:
                self.log.warn("Unable to save environment: %s",
                              stacktrace.str_unpickable_object(env.data))
            else:
                self.log.warn("Unable to save environment: %s (%s)", details,
                              env.data)
            return True
        return False

    def setUp(self):
        """
        Avocado-vt uses custom setUp/test/tearDown handling and unlike
        Avocado it allows skipping tests from any phase. To convince
        Avocado to allow skips let's say our tests run during setUp
        phase and report the status in test.
        """
        env_lang = os.environ.get('LANG')
        os.environ['LANG'] = 'C'
        try:
            self._runTest()
            self.__status = "PASS"
        # This trick will give better reporting of virt tests being executed
        # into avocado (skips, warns and errors will display correctly)
        except exceptions.TestSkipError:
            self.__exc_info = sys.exc_info()
            raise   # This one has to be raised in setUp
        except:  # nopep8 Old-style exceptions are not inherited from Exception()
            self.__exc_info = sys.exc_info()
            self.__status = self.__exc_info[1]
        finally:
            # Clean libvirtd debug logs if the test is not fail or error
            if self.params.get("libvirtd_log_cleanup", "no") == "yes":
                if(self.params.get("vm_type") == 'libvirt' and
                   self.params.get("enable_libvirtd_debug_log", "yes") == "yes"):
                    libvirtd_log = self.params["libvirtd_debug_file"]
                    if("TestFail" not in str(self.__exc_info) and
                       "TestError" not in str(self.__exc_info)):
                        if libvirtd_log and os.path.isfile(libvirtd_log):
                            logging.info("cleaning libvirtd logs...")
                            os.remove(libvirtd_log)
                    else:
                        # tar the libvirtd log and archive
                        logging.info("archiving libvirtd debug logs")
                        from virttest import utils_package
                        if utils_package.package_install("tar"):
                            if os.path.isfile(libvirtd_log):
                                archive = os.path.join(os.path.dirname(
                                    libvirtd_log), "libvirtd.tar.gz")
                                cmd = ("tar -zcf %s -P %s"
                                       % (pipes.quote(archive),
                                          pipes.quote(libvirtd_log)))
                                if process.system(cmd) == 0:
                                    os.remove(libvirtd_log)
                            else:
                                logging.error("Unable to find log file: %s",
                                              libvirtd_log)
                        else:
                            logging.error("Unable to find tar to compress libvirtd "
                                          "logs")

            if env_lang:
                os.environ['LANG'] = env_lang
            else:
                del os.environ['LANG']

    def runTest(self):
        """
        This only reports the results

        The actual testing happens inside setUp stage, this only
        reports the correct results
        """
        if self.__status != "PASS":
            raise self.__status  # pylint: disable=E0702

    def _runTest(self):
        params = self.params

        # If a dependency test prior to this test has failed, let's fail
        # it right away as TestNA.
        if params.get("dependency_failed") == 'yes':
            raise exceptions.TestSkipError("Test dependency failed")

        # Report virt test version
        logging.info(version.get_pretty_version_info())
        # Report the parameters we've received and write them as keyvals
        logging.debug("Test parameters:")
        keys = list(params.keys())
        keys.sort()
        for key in keys:
            logging.debug("    %s = %s", key, params[key])

        # Warn of this special condition in related location in output & logs
        if os.getuid() == 0 and params.get('nettype', 'user') == 'user':
            logging.warning("")
            logging.warning("Testing with nettype='user' while running "
                            "as root may produce unexpected results!!!")
            logging.warning("")

        # Find the test
        subtest_dirs = []
        test_filter = bootstrap.test_filter

        other_subtests_dirs = params.get("other_tests_dirs", "")
        for d in other_subtests_dirs.split():
            # If d starts with a "/" an absolute path will be assumed
            # else the relative path will be searched in the bin_dir
            subtestdir = os.path.join(self.bindir, d, "tests")
            if not os.path.isdir(subtestdir):
                raise exceptions.TestError("Directory %s does not "
                                           "exist" % subtestdir)
            subtest_dirs += data_dir.SubdirList(subtestdir,
                                                test_filter)

        provider = params.get("provider", None)

        if provider is None:
            # Verify if we have the correspondent source file for
            # it
            generic_subdirs = asset.get_test_provider_subdirs(
                'generic')
            for generic_subdir in generic_subdirs:
                subtest_dirs += data_dir.SubdirList(generic_subdir,
                                                    test_filter)
            specific_subdirs = asset.get_test_provider_subdirs(
                params.get("vm_type"))
            for specific_subdir in specific_subdirs:
                subtest_dirs += data_dir.SubdirList(
                    specific_subdir, bootstrap.test_filter)
        else:
            provider_info = asset.get_test_provider_info(provider)
            for key in provider_info['backends']:
                subtest_dirs += data_dir.SubdirList(
                    provider_info['backends'][key]['path'],
                    bootstrap.test_filter)

        subtest_dir = None

        # Get the test routine corresponding to the specified
        # test type
        logging.debug("Searching for test modules that match "
                      "'type = %s' and 'provider = %s' "
                      "on this cartesian dict",
                      params.get("type"),
                      params.get("provider", None))

        t_types = params.get("type").split()
        # Make sure we can load provider_lib in tests
        for s in subtest_dirs:
            if os.path.dirname(s) not in sys.path:
                sys.path.insert(0, os.path.dirname(s))

        test_modules = {}
        for t_type in t_types:
            for d in subtest_dirs:
                module_path = os.path.join(d, "%s.py" % t_type)
                if os.path.isfile(module_path):
                    logging.debug("Found subtest module %s",
                                  module_path)
                    subtest_dir = d
                    break
            if subtest_dir is None:
                msg = ("Could not find test file %s.py on test"
                       "dirs %s" % (t_type, subtest_dirs))
                raise exceptions.TestError(msg)
            # Load the test module
            f, p, d = imp.find_module(t_type, [subtest_dir])
            test_modules[t_type] = imp.load_module(t_type, f, p, d)
            f.close()

        # Open the environment file
        env_filename = os.path.join(data_dir.get_tmp_dir(),
                                    params.get("env", "env"))
        env = utils_env.Env(env_filename, self.env_version)
        if params.get_boolean("job_env_cleanup", "yes"):
            self.runner_queue.put({"func_at_exit": cleanup_env,
                                   "args": (env_filename, self.env_version),
                                   "once": True})

        test_passed = False
        t_type = None

        try:
            try:
                try:
                    # Pre-process
                    try:
                        params = env_process.preprocess(self, params, env)
                    finally:
                        self.__safe_env_save(env)

                    # Run the test function
                    for t_type in t_types:
                        test_module = test_modules[t_type]
                        run_func = utils_misc.get_test_entrypoint_func(
                            t_type, test_module)
                        try:
                            run_func(self, params, env)
                            self.verify_background_errors()
                        finally:
                            self.__safe_env_save(env)
                    test_passed = True
                    error_message = funcatexit.run_exitfuncs(env, t_type)
                    if error_message:
                        raise exceptions.TestWarn("funcatexit failed with: %s" %
                                                  error_message)

                except:  # nopep8 Old-style exceptions are not inherited from Exception()
                    stacktrace.log_exc_info(sys.exc_info(), 'avocado.test')
                    if t_type is not None:
                        error_message = funcatexit.run_exitfuncs(env, t_type)
                        if error_message:
                            logging.error(error_message)
                    try:
                        env_process.postprocess_on_error(self, params, env)
                    finally:
                        self.__safe_env_save(env)
                    raise

            finally:
                # Post-process
                try:
                    try:
                        params['test_passed'] = str(test_passed)
                        env_process.postprocess(self, params, env)
                    except:  # nopep8 Old-style exceptions are not inherited from Exception()

                        stacktrace.log_exc_info(sys.exc_info(),
                                                'avocado.test')
                        if test_passed:
                            raise
                        logging.error("Exception raised during "
                                      "postprocessing: %s",
                                      sys.exc_info()[1])
                finally:
                    if self.__safe_env_save(env) or params.get("env_cleanup", "no") == "yes":
                        env.destroy()   # Force-clean as it can't be stored

        except Exception as e:
            if params.get("abort_on_error") != "yes":
                raise
            # Abort on error
            logging.info("Aborting job (%s)", e)
            if params.get("vm_type") == "qemu":
                for vm in env.get_all_vms():
                    if vm.is_dead():
                        continue
                    logging.info("VM '%s' is alive.", vm.name)
                    for m in vm.monitors:
                        logging.info("It has a %s monitor unix socket at: %s",
                                     m.protocol, m.filename)
                    logging.info("The command line used to start it was:\n%s",
                                 vm.make_create_command())
                raise exceptions.JobError("Abort requested (%s)" % e)

        return test_passed



"""
tools to manage sasl.
"""

import logging

import aexpect
from aexpect import remote

from avocado.core import exceptions
from avocado.utils import path
from avocado.utils import process

from virttest import propcan
from virttest import virsh


class SASL(propcan.PropCanBase):

    """
    Base class of a connection between server and client.
    """
    __slots__ = ("sasl_pwd_cmd", "sasl_user_pwd", "sasl_user_cmd",
                 "auto_recover", "linesep", "prompt", "session",
                 "server_ip", "server_user", "server_pwd",
                 "client", "port")

    def __init__(self, *args, **dargs):
        """
        Initialize instance
        """
        init_dict = dict(*args, **dargs)
        init_dict["sasl_pwd_cmd"] = path.find_command("saslpasswd2")
        init_dict["sasl_user_cmd"] = path.find_command("sasldblistusers2")
        init_dict["sasl_user_pwd"] = init_dict.get("sasl_user_pwd")
        init_dict["auto_recover"] = init_dict.get("auto_recover", False)
        init_dict["client"] = init_dict.get("client", "ssh")
        init_dict["port"] = init_dict.get("port", "22")
        init_dict["linesep"] = init_dict.get("linesep", "\n")
        init_dict["prompt"] = init_dict.get("prompt", r"[\#\$]\s*$")

        self.__dict_set__('session', None)
        super(SASL, self).__init__(init_dict)

    def __del__(self):
        """
        Close opened session and clear test environment
        """
        self.close_session()
        if self.auto_recover:
            try:
                self.cleanup()
            except Exception:
                raise exceptions.TestError(
                    "Failed to clean up test environment!")

    def _new_session(self):
        """
        Build a new server session.
        """
        port = self.port
        prompt = self.prompt
        host = self.server_ip
        client = self.client
        username = self.server_user
        password = self.server_pwd

        try:
            session = remote.wait_for_login(client, host, port,
                                            username, password, prompt)
        except remote.LoginTimeoutError:
            raise exceptions.TestError(
                "Got a timeout error when login to server.")
        except remote.LoginAuthenticationError:
            raise exceptions.TestError(
                "Authentication failed to login to server.")
        except remote.LoginProcessTerminatedError:
            raise exceptions.TestError(
                "Host terminates during login to server.")
        except remote.LoginError:
            raise exceptions.TestError(
                "Some error occurs login to client server.")
        return session

    def get_session(self):
        """
        Make sure the session is alive and available
        """
        session = self.__dict_get__('session')

        if (session is not None) and (session.is_alive()):
            return session
        else:
            session = self._new_session()

        self.__dict_set__('session', session)
        return session

    def close_session(self):
        """
        If session exists then close it
        """
        if self.session:
            self.session.close()

    def list_users(self, remote=True, sasldb_path="/etc/libvirt/passwd.db"):
        """
        List users in sasldb
        """
        cmd = "%s -f %s" % (self.sasl_user_cmd, sasldb_path)
        try:
            if remote:
                self.session = self.get_session()
                return self.session.cmd_output(cmd)
            else:
                return process.run(cmd).stdout_text
        except process.CmdError:
            logging.error("Failed to set a user's sasl password %s", cmd)

    def setup(self, remote=True):
        """
        Create sasl users with password
        """
        for sasl_user, sasl_pwd in eval(self.sasl_user_pwd):
            cmd = "echo %s |%s -p -a libvirt %s" % (sasl_pwd,
                                                    self.sasl_pwd_cmd,
                                                    sasl_user)
            try:
                if remote:
                    self.session = self.get_session()
                    self.session.cmd(cmd)
                else:
                    process.system(cmd)
            except process.CmdError:
                logging.error("Failed to set a user's sasl password %s", cmd)

    def cleanup(self, remote=True):
        """
        Clear created sasl users
        """
        for sasl_user, sasl_pwd in eval(self.sasl_user_pwd):
            cmd = "%s -a libvirt -d %s" % (self.sasl_pwd_cmd, sasl_user)
            try:
                if remote:
                    self.session = self.get_session()
                    self.session.cmd(cmd)
                else:
                    process.system(cmd)
            except process.CmdError:
                logging.error("Failed to disable a user's access %s", cmd)


class VirshSessionSASL(virsh.VirshSession):

    """
    A wrap class for virsh session which used SASL infrastructure.
    """

    def __init__(self, params):
        self.virsh_exec = virsh.VIRSH_EXEC
        self.sasl_user = params.get('sasl_user')
        self.sasl_pwd = params.get('sasl_pwd')
        self.remote_ip = params.get('remote_ip')
        self.remote_user = params.get('remote_user')
        self.remote_pwd = params.get('remote_pwd')
        self.remote_auth = False
        if self.remote_ip:
            self.remote_auth = True
        super(VirshSessionSASL, self).__init__(virsh_exec=self.virsh_exec,
                                               remote_ip=self.remote_ip,
                                               remote_user=self.remote_user,
                                               remote_pwd=self.remote_pwd,
                                               ssh_remote_auth=self.remote_auth,
                                               auto_close=True,
                                               check_libvirtd=False)
        self.sendline('connect')
        self.sendline(self.sasl_user)
        self.sendline(self.sasl_pwd)
        # make sure session is connected successfully
        if self.cmd_status('list', timeout=60) != 0:
            logging.debug("Persistent virsh session is not responding, "
                          "libvirtd may be dead.")
            raise aexpect.ShellStatusError(virsh.VIRSH_EXEC, 'list')
